# -*- coding: utf-8 -*-
__all__ = ('Achievement', 'OA2Access', 'UserOA2', 'parse_oauth2_redirect_url')

import re
from datetime import datetime
from time import time as time_now

from .bases import DiscordEntity, IconSlot
from .integration import Integration
from .utils import DATETIME_FORMAT_CODE
from .user import UserBase, UserFlag
from .preinstanced import PremiumType

from . import urls as module_urls

DEFAULT_LOCALE = 'en-US'
LOCALES = {DEFAULT_LOCALE: DEFAULT_LOCALE}

def parse_locale(data):
    """
    Gets `'local'`'s value out from the given `dict`. If found returns it, if not, then returns `DEFAULT_LOCAL`.
    
    To not keep using new local values at every case, the already used local values are cached at `LOCALE`.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        Some data received from Discord.
    
    Returns
    -------
    locale : `str`
    """
    try:
        locale = data['locale']
    except KeyError:
        return DEFAULT_LOCALE
    
    locale = LOCALES.setdefault(locale, locale)
    return locale

def parse_preferred_locale(data):
    """
    Gets `'preferred_locale'`'s value out from the given `dict`. If found returns it, if not, then returns `DEFAULT_LOCAL`.
    
    To not keep using new local values at every case, the already used local values are cached at `LOCALE`.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        Some data received from Discord.
    
    Returns
    -------
    locale : `str`
    """
    try:
        locale = data['preferred_locale']
    except KeyError:
        return DEFAULT_LOCALE
    
    locale = LOCALES.setdefault(locale, locale)
    return locale

def parse_locale_optional(data):
    """
    Gets `'local'`'s value out from the given `dict`. If found returns it, if not, then returns `None`.
    
    To not keep using new local values at every case, the already used local values are cached at `LOCALE`.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        Some data received from Discord.
    
    Returns
    -------
    locale : `str` or `None`
    """
    try:
        locale = data['locale']
    except KeyError:
        return None
    
    locale = LOCALES.setdefault(locale, locale)
    return locale

OA2_RU_RP = re.compile('(https?://.+?)\?code=([a-zA-Z0-9]{30})')

def parse_oauth2_redirect_url(url):
    """
    Parses the `redirect_url` and the `code` out from a whole `url`, what an user was redirected to after oauth2
    authorization.
    
    If the parsing was successful, then returns a `tuple` of `redirect_url` and `code`. If it fails, returns `None`.
    
    Parameters
    ----------
    url : `str`
        A whole to url to parse from

    Returns
    -------
    result : `None` or `tuple` (`str`, `str`)
    """
    result = OA2_RU_RP.fullmatch(url)
    if result is None:
        return None
    
    return result.groups()

class Connection(DiscordEntity):
    """
    A connection object that a user is attached to.
    
    Attributes
    ----------
    id : `int`
        The unique identifier value of the connection.
    friend_sync : `bool`
        Whether the user has friend sync enabled for the connection.
    integrations : `None` or (`list` of ``Integration``)
        A list (if any) of guild integrations which are attached to the connection.
    name : `str`
        The username of the connected account.
    revoked : `bool`
        Whether the connection is revoked.
    show_activity : `bool`
        Whether activity related to this connection will be shown in presence updates.
    type : `str`
        The service of the connection. (Like `'twitch'` or `'youtube'`.)
    verified : `bool`
        Whether the connection is verified.
    visibility : `int`
        For who is the connection visible for.
        
        Possible visibility values
        +-------+-------------------------------+
        | value | description                   |
        +=======+===============================+
        | 0     | Visible only for the user.    |
        +-------+-------------------------------+
        | 1     | Visible to everyone.          |
        +-------+-------------------------------+
    """
    __slots__ = ('friend_sync', 'integrations', 'name', 'revoked', 'show_activity', 'type', 'verified', 'visibility',)
    
    def __init__(self, data):
        """
        Creates a connection object from received connection data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`)
            Received connection data.
        """
        self.name = data['name']
        self.type = data['type']
        self.id = int(data['id'])
        self.revoked = data.get('revoked', False)
        self.verified = data.get('verified', False)
        self.show_activity = data.get('show_activity', False)
        self.friend_sync = data.get('friend_sync', False)
        self.visibility = data.get('visibility', 0)
        
        try:
            integration_datas = data['integrations']
        except KeyError:
            integrations = None
        else:
            if integration_datas:
                integrations = [Integration(integration_data) for integration_data in integration_datas]
            else:
                integrations = None
        self.integrations = integrations


SCOPES = {v: v for v in ('activities.read', 'activities.write', 'applications.builds.read',
    'applications.builds.upload', 'applications.commands', 'applications.entitlements', 'applications.store.update',
    'bot', 'connections', 'email', 'guilds', 'guilds.join', 'identify', 'messages.read', 'rpc', 'rpc.api',
    'rpc.notifications.read', 'webhook.incoming')}

# rest of scopes are ignorable

class OA2Access:
    """
    Represents a Discord oauth2 access object, what is returned by ``Client.activate_authorization_code`` if
    activating the authorization code went successfully.
    
    Attributes
    ----------
    access_token : `str`
        Token used for `Bearer` authorizations, when requesting OAuth2 data about the respective user.
    created_at : `datetime`
        The time when the access was last created or renewed.
    expires_in : `int`
        The time in seconds after this access expires.
    redirect_url : `str`
        The redirect url with what the user granted the authorization code for the oauth2 scopes for the application.
        
        Can be empty string if application's owner's access was requested.
    refresh_token : `str`
        The token used to renew the access token.
        
        Can be empty string if application's owner's access was requested.
    scopes : `set` of `str`
        A list of the scopes, what the user granted with the access token.
    
    Class Attributes
    ----------------
    TOKEN_TYPE : `str` = `'Bearer'`
        The access token's type.
    """
    TOKEN_TYPE = 'Bearer'
    
    __slots__ = ('access_token', 'created_at', 'expires_in', 'redirect_url', 'refresh_token', 'scopes',)
    def __init__(self, data, redirect_url):
        """
        Creates an ``OA2Access``.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received access data.
        redirect_url : `str`
            The redirect url with what the user granted the authorization code for the oauth2 scopes for the
            application.
        """
        self.redirect_url = redirect_url
        self.access_token = data['access_token']
        self.refresh_token = data.get('refresh_token', '')
        self.expires_in = data['expires_in'] # default is 604800 (s) (1 week)
        self.scopes = scopes = set()
        for scope in data['scope'].split():
            try:
                scopes.add(SCOPES[scope])
            except KeyError:
                pass
        self.created_at = datetime.utcnow() #important for renewing
    
    def _renew(self, data):
        """
        Renews the access with the given data.
        
        Parameters
        ----------
        data : `None` or (`dict` of (`str`, `Any`))
            Requested access data.
        """
        self.created_at = datetime.utcnow()
        if data is None:
            return
        
        self.access_token = data['access_token']
        self.refresh_token = data.get('refresh_token', '')
        self.expires_in = data['expires_in']
        scopes = self.scopes
        scopes.clear()
        for scope in data['scope'].split():
            try:
                scopes.add(SCOPES[scope])
            except KeyError:
                pass
    
    def __repr__(self):
        """Returns the representation of the achievement."""
        state = 'active' if (self.created_at.timestamp()+self.expires_in > time_now()) else 'expired'
        return (f'<{self.__class__.__name__} {state}, access_token={self.access_token!r}, scopes count='
            f'{len(self.scopes)}>')


class UserOA2(UserBase):
    """
    Represents a Discord user with extra personal information. If a ``UserOA2`` is  created it will NOT overwrite the
    already existing user with the same ID, if exists.
    
    Attributes
    ----------

    id : `int`
        The user's unique identifier number.
    name : str
        The user's username.
    discriminator : `int`
        The user's discriminator. Given to avoid overlapping names.
    avatar_hash : `int`
        The user's avatar's hash in `uint128`.
    avatar_type : `bool`
        The user's avatar's type.
    email : `None` or `str`
        The user's email. Defaults to empty string.
    flags : ``UserFlag``
        The user's user flags.
    locale : `str`
        The preferred locale by the user.
    mfa : `bool`
        Whether the user has two factor authorization enabled on the account.
    premium_type : ``PremiumType``
        The Nitro subscription type of the user.
    system : `bool`
        Whether the user is an Official Discord System user (part of the urgent message system).
    verified : `bool`
        Whether the email of the user is verified.
    """
    __slots__ = ('access', 'email', 'flags', 'locale', 'mfa', 'premium_type', 'system', 'verified', )
    
    def __init__(self, data, access):
        self.access = access
        self.id = int(data['id'])
        self.name = data['username']
        self.discriminator = int(data['discriminator'])
        
        self._set_avatar(data)
        
        self.mfa = data.get('mfa_enabled', False)
        self.verified = data.get('verified', False)
        self.email = data.get('email', None)
        
        try:
            flags = data['flags']
        except KeyError:
            flags = data.get('public_flags', 0)
        
        self.flags = UserFlag(flags)
        self.premium_type = PremiumType.get(data.get('premium_type', 0))
        self.locale = parse_locale(data)
        self.system = data.get('system', False)
    
    @property
    def partial(self):
        """
        Returns whether the oauth2 user object is partial.
        
        Returns
        -------
        partial : `bool` = `False`
        """
        return False
    
    @property
    def is_bot(self):
        """
        Returns whether the oauth2 user represents a bot account.
        
        Returns
        -------
        is_bot : `bool` = `False`
        """
        return False
    
    # Reflect OA2Access
    @property
    def access_token(self):
        """
        Returns the oauth2 user's access's token.
        
        Returns
        -------
        access_token : `str`
        """
        return self.access.access_token
    
    @property
    def redirect_url(self):
        """
        Returns the oauth2 user's access's redirect url.
        
        Returns
        -------
        redirect_url : `str`
        """
        return self.access.redirect_url
    
    @property
    def refresh_token(self):
        """
        Returns the oauth2 user's access's refresh token.
        
        Returns
        -------
        refresh_token : `str`
        """
        return self.access.refresh_token
    
    @property
    def scopes(self):
        """
        Returns the oauth2 user's access's scopes.
        
        Returns
        -------
        scopes : `set` of `str`
        """
        return self.access.scopes

    def _renew(self, data):
        """
        Renews the oauth2 user's access with the given data.
        
        Parameters
        ----------
        data : `None` or (`dict` of (`str`, `Any`))
            Requested access data.
        """
        self.access._renew(data)


class Achievement(DiscordEntity):
    """
    Represents a Discord achievement created at Developer portal.
    
    Attributes
    ----------
    id : `int`
        The achievement's unique identifier number.
    application_id : `int`
        The achievement's respective application's id.
    description : `str`
        The description of the achievement.
    name : `str`
        The name of the achievement.
    secret : `bool`
        Secret achievements will *not* be shown to the user until they've unlocked them.
    secure : `bool`
        Secure achievements can only be set via HTTP calls from your server, not by a game client using the SDK.
    icon_hash : `int`
        The achievement's icon's hash. Achievements always have icon.
    icon_type : ``IconType``
        The achievement's icon's type.
    """
    __slots__ = ('application_id', 'description', 'name', 'secret', 'secure',)
    
    icon = IconSlot('icon', 'icon_hash', module_urls.achievement_icon_url, module_urls.achievement_icon_url_as)
    
    def __init__(self, data):
        """
        Creates an achievement with the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received achievement data.
        """
        self.application_id = int(data['application_id'])
        self.id = int(data['id'])
        
        self._update_no_return(data)
    
    def __repr__(self):
        """Returns the achievement's representation."""
        return f'<{self.__class__.__name__} name={self.name!r}, id={self.id}>'
    
    def __str__(self):
        """Returns the achievement's name."""
        return self.name
    
    def __format__(self,code):
        if not code:
            return self.name
        
        if code=='c':
            return self.created_at.__format__(DATETIME_FORMAT_CODE)
        
        raise ValueError(f'Unknown format code {code!r} for object of type {self.__class__.__name__!r}')
    
    def _update(self, data):
        """
        Updates the achievement and returns it's overwritten attributes as a `dict` with a `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Achievement data received from Discord.
            
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        +---------------+-----------+
        | Keys          | Values    |
        +===============+===========+
        | name          | `str`     |
        +---------------+-----------+
        | description   | `str`     |
        +---------------+-----------+
        | secret        | `bool`    |
        +---------------+-----------+
        | secure        | `bool`    |
        +---------------+-----------+
        | icon          | ``Icon``  |
        +---------------+-----------+
        """
        old_attributes = {}
        
        name = data['name']['default']
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        description = data['description']['default']
        if self.description != description:
            old_attributes['description'] = self.description
            self.description = description
        
        secret = data['secret']
        if self.secret != secret:
            old_attributes['secret'] = self.secret
            self.secret = secret
        
        secure = data['secure']
        if self.secure != secure:
            old_attributes['secure'] = self.secure
            self.secure = secure
        
        self._update_icon(data, old_attributes)
        
        return old_attributes
    
    def _update_no_return(self, data):
        """
        Updates the achievement with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Achievement data received from Discord.
        """
        self.name = data['name']['default']
        self.description = data['description']['default']
        
        self.secret = data['secret']
        self.secure = data['secure']
        
        self._set_icon(data)



# -*- coding: utf-8 -*-
__all__ = ('SlashResponse', 'abort', )
from ...backend.futures import is_coroutine_generator

from ...discord.parsers import InteractionEvent, INTERACTION_EVENT_RESPONSE_STATE_NONE, \
    INTERACTION_EVENT_RESPONSE_STATE_DEFERRED, INTERACTION_EVENT_RESPONSE_STATE_RESPONDED
from ...discord.exceptions import DiscordException, ERROR_CODES
from ...discord.embed import EmbedBase
from ...discord.client import Client

def is_only_embed(maybe_embeds):
    """
    Checks whether the given value is a `tuple` or `list` containing only `embed-like`-s.
    
    Parameters
    ----------
    maybe_embeds : (`tuple` or `list`) of `EmbedBase` or `Any`
        The value to check whether is a `tuple` or `list` containing only `embed-like`-s.
    
    Returns
    -------
    is_only_embed : `bool`
    """
    if not isinstance(maybe_embeds, (list, tuple)):
        return False
    
    for maybe_embed in maybe_embeds:
        if not isinstance(maybe_embed, EmbedBase):
            return False
    
    return True

async def get_request_coros(client, interaction_event, show_for_invoking_user_only, response):
    """
    Gets request coroutine after an output from a command coroutine. Might return `None` if there is nothing to send.
    
    This function is a coroutine generator, which should be ued inside of an async for loop.
    
    Parameters
    ----------
    client : ``Client``
        The client who will send the responses if applicable.
    interaction_event : ``InteractionEvent``
        The respective event to respond on.
    show_for_invoking_user_only : `bool`
        Whether the response message should only be shown for the invoking user.
    response : `Any`
        Any object yielded or returned by the command coroutine.
    
    Yields
    -------
    request_coro : `None` or `coroutine`
    """
    response_state = interaction_event._response_state
    if (response is None):
        if response_state == INTERACTION_EVENT_RESPONSE_STATE_NONE:
            yield client.interaction_response_message_create(interaction_event,
                show_for_invoking_user_only=show_for_invoking_user_only)
        
        return
    
    if isinstance(response, (str, EmbedBase)) or is_only_embed(response):
        if response_state == INTERACTION_EVENT_RESPONSE_STATE_NONE:
            yield client.interaction_response_message_create(interaction_event, response,
                show_for_invoking_user_only=show_for_invoking_user_only)
            return
        
        if response_state == INTERACTION_EVENT_RESPONSE_STATE_DEFERRED:
            yield client.interaction_response_message_edit(interaction_event, response)
            return
        
        if response_state == INTERACTION_EVENT_RESPONSE_STATE_RESPONDED:
            yield client.interaction_followup_message_create(interaction_event, response,
                show_for_invoking_user_only=show_for_invoking_user_only)
            return
        
        # No more cases
        return
        
    if is_coroutine_generator(response):
        response = await process_command_gen(client, interaction_event, show_for_invoking_user_only, response)
        async for request_coro in get_request_coros(client, interaction_event, show_for_invoking_user_only, response):
            yield request_coro
        
        return
    
    if isinstance(response, SlashResponse):
        for request_coro in response.get_request_coros(client, interaction_event, show_for_invoking_user_only):
            yield request_coro
        
        return
    
    if response_state == INTERACTION_EVENT_RESPONSE_STATE_NONE:
        yield client.interaction_response_message_create(interaction_event,
            show_for_invoking_user_only=show_for_invoking_user_only)
        
        return
    
    # No more cases
    return


async def process_command_gen(client, interaction_event, show_for_invoking_user_only, coro):
    """
    Processes a slash command coroutine generator.
    
    This function os a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The client who will send the responses if applicable.
    interaction_event : ``InteractionEvent``
        The respective event to respond on.
    show_for_invoking_user_only : `bool`
        Whether the response message should only be shown for the invoking user.
    coro : `CoroutineGenerator`
        A coroutine generator with will send command response.
    
    Returns
    -------
    response : `Any`
        Returned object by the coroutine generator.
    
    Raises
    ------
    BaseException
        Any exception raised by `coro`.
    """
    response_message = None
    response_exception = None
    while True:
        if response_exception is None:
            step = coro.asend(response_message)
        else:
            step = coro.athrow(response_exception)
        
        try:
            response = await step
        except StopAsyncIteration as err:
            # catch `StopAsyncIteration` only if it is a new one.
            if (response_exception is not None) and (response_exception is not err):
                raise
            
            args = err.args
            if args:
                response = args[0]
            else:
                response = None
            break
        
        except InteractionAbortedError as err:
            response = err.response
            break
        
        except BaseException as err:
            if (response_exception is None) or (response_exception is not err):
                raise
            
            if isinstance(err, ConnectionError):
                return
            
            if isinstance(err, DiscordException):
                if err.code in (
                        ERROR_CODES.unknown_channel, # Message's channel deleted; Can we get this?
                        ERROR_CODES.invalid_access, # Client removed.
                        ERROR_CODES.invalid_permissions, # Permissions changed meanwhile; Can we get this?
                        ERROR_CODES.cannot_message_user, # User has dm-s disallowed; Can we get this?
                        ERROR_CODES.unknown_interaction, # We times out, do not drop error.
                            ):
                    return
            
            raise
        
        else:
            # We set it first, since if `get_request_coros` yields nothing, we would be meowed up.
            response_message = None
            response_exception = None
            
            async for request_coro in get_request_coros(client, interaction_event, show_for_invoking_user_only,
                    response):
                try:
                    response_message = await request_coro
                except BaseException as err:
                    # `response_message` may have be set before with an iteration, so reset it.
                    response_message = None
                    response_exception = err
                    break
    
    
    return response


async def process_command_coro(client, interaction_event, show_for_invoking_user_only, coro):
    """
    Processes a slash command coroutine.
    
    If the coroutine returns or yields a string or an embed like then sends it to the respective channel.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The client who will send the responses if applicable.
    interaction_event : ``InteractionEvent``
        The respective event to respond on.
    show_for_invoking_user_only : `bool`
        Whether the response message should only be shown for the invoking user.
    coro : `coroutine`
        A coroutine with will send command response.
    
    Raises
    ------
    BaseException
        Any exception raised by `coro`.
    """
    if is_coroutine_generator(coro):
        response = await process_command_gen(client, interaction_event, show_for_invoking_user_only, coro)
    else:
        try:
            response = await coro
        except InteractionAbortedError as err:
            response = err.response
    
    async for request_coro in get_request_coros(client, interaction_event, show_for_invoking_user_only, response):
        try:
            await request_coro
        except BaseException as err:
            if isinstance(err, ConnectionError):
                return
            
            if isinstance(err, DiscordException):
                if err.code in (
                        ERROR_CODES.unknown_channel, # message's channel deleted; Can we get this?
                        ERROR_CODES.invalid_access, # client removed.
                        ERROR_CODES.invalid_permissions, # permissions changed meanwhile; Can we get this?
                        ERROR_CODES.cannot_message_user, # user has dm-s disallowed; Can we get this?
                        ERROR_CODES.unknown_interaction, # we timed out, do not drop error.
                            ):
                    return
            
            raise


class SlashResponse:
    """
    Rich interaction response message usable with `return` or with `yield` statements.
    
    May pass it's parameters to any oft he following method depending on control flow.
    
    - ``Client.interaction_response_message_create``
    - ``Client.interaction_response_message_edit``
    - ``Client.interaction_followup_message_create``
    
    Attributes
    ----------
    -force_new_message : `int`
        Whether a new message should be forced out from Discord and being retrieved.
    _parameters : `dict` of (`str`, `Any`) items
        Parameters to pass to the respective ``Client`` functions.
        
        Can have the following keys:
        
        - `'allowed_mentions'`
        - `'content'`
        - `'components'`
        - `'embed'`
        - `'file'`
        - `'show_for_invoking_user_only'`
        - `'tts'`
    """
    __slots__ = ('_force_new_message', '_parameters',)
    
    def __init__(self, content=..., *, embed=..., file=..., allowed_mentions=..., tts=..., components=None,
            show_for_invoking_user_only=..., force_new_message=0):
        """
        Creates a new ``SlashResponse`` instance with the given parameters.
        
        Parameters
        ----------
        content : `str`, ``EmbedBase``, `Any`, Optional
            The message's content if given. If given as `str` or empty string, then no content will be sent, meanwhile
            if any other non `str` or ``EmbedBase`` instance is given, then will be casted to string.
            
            If given as ``EmbedBase`` instance, then is sent as the message's embed.
            
        embed : ``EmbedBase`` instance or `list` of ``EmbedBase`` instances, Optional (Keyword only)
            The embedded content of the message.
            
            If `embed` and `content` parameters are both given as  ``EmbedBase`` instance, then `TypeError` is raised.
        file : `Any`, Optional (Keyword only)
            A file to send. Check ``Client._create_file_form`` for details.
        allowed_mentions : `None`,  `str`, ``UserBase``, ``Role``, `list` of (`str`, ``UserBase``, ``Role`` )
                , Optional (Keyword only)
            Which user or role can the message ping (or everyone). Check ``Client._parse_allowed_mentions`` for details.
        components : `None`, ``ComponentBase``, (`set`, `list`) of ``ComponentBase``, Optional (Keyword only)
            Components attached to the message.
        tts : `bool`, Optional (Keyword only)
            Whether the message is text-to-speech.
        show_for_invoking_user_only : `bool`, Optional (Keyword only)
            Whether the sent message should only be shown to the invoking user. Defaults to the value passed when adding
            the command.
            
            If given as `True` only the message's content will be processed by Discord.
        force_new_message : `int` or `bool`, Optional (Keyword only)
            Whether a new message should be forced out from Discord allowing the client to retrieve a new ``Message``
            object as well. Defaults to `False`.
            
            If given as `-1` will only force new message if the event already deferred.
        """
        self._force_new_message = force_new_message
        self._parameters = parameters = {}
        
        if (content is not ...):
            parameters['content'] = content
        
        if (embed is not ...):
            parameters['embed'] = embed
        
        if (file is not ...):
            parameters['file'] = file
        
        if (allowed_mentions is not ...):
            parameters['allowed_mentions'] = allowed_mentions
        
        if (components is not ...):
            parameters['components'] = components
        
        if (tts is not ...):
            parameters['tts'] = tts
        
        if (show_for_invoking_user_only is not ...):
            parameters['show_for_invoking_user_only'] = show_for_invoking_user_only
    
    def _get_response_parameters(self, allowed_parameters):
        """
        Gets response parameters to pass to a ``Client`` method.
        
        Parameters
        ----------
        allowed_parameters : `tuple` of `str`
            Allowed parameters to be passed to the respective client method.
        
        Returns
        -------
        response_parameters : `dict` of (`str`, `Any`) items
            Parameters to pass the the respective client method.
        """
        parameters = self._parameters
        response_parameters = {}
        for key in allowed_parameters:
            try:
                value = parameters[key]
            except KeyError:
                continue
            
            response_parameters[key] = value
        
        return response_parameters
    
    def get_request_coros(self, client, interaction_event, show_for_invoking_user_only):
        """
        Gets request coroutine buildable from the ``SlashResponse``.
        
        This method is a generator, which should be used inside of a `for` loop.
        
        client : ``Client``
            The client who will send the responses if applicable.
        interaction_event : ``InteractionEvent``
            The respective event to respond on.
        show_for_invoking_user_only : `bool`
            Whether the response message should only be shown for the invoking user.
        
        Yields
        -------
        request_coro : `None` or `coroutine`
        """
        response_state = interaction_event._response_state
        force_new_message = self._force_new_message
        if force_new_message:
            if force_new_message > 0:
                if response_state == INTERACTION_EVENT_RESPONSE_STATE_NONE:
                    yield client.interaction_response_message_create(interaction_event)
                
                response_parameters = self._get_response_parameters(('allowed_mentions', 'content', 'embed', 'file',
                    'tts', 'components',))
                response_parameters['show_for_invoking_user_only'] = \
                    self._parameters.get('show_for_invoking_user_only', show_for_invoking_user_only)
                
                yield client.interaction_followup_message_create(interaction_event, **response_parameters)
                return
            else:
                if response_state == INTERACTION_EVENT_RESPONSE_STATE_NONE:
                    if 'file' in self._parameters:
                        show_for_invoking_user_only = \
                            self._parameters.get('show_for_invoking_user_only', show_for_invoking_user_only)
                        
                        yield client.interaction_response_message_create(interaction_event,
                            show_for_invoking_user_only=show_for_invoking_user_only)
                        
                        response_parameters = self._get_response_parameters(('allowed_mentions', 'content', 'embed',
                            'file', 'tts', 'components'))
                        
                        if 'components' in response_parameters:
                            # Enforce new message if components is present. They are not processed by interaction
                            # message edit events.
                            yield client.interaction_followup_message_create(interaction_event,
                                **response_parameters)
                        else:
                            yield client.interaction_response_message_edit(interaction_event,
                                **response_parameters)
                    else:
                        response_parameters = self._get_response_parameters(('allowed_mentions', 'content', 'embed',
                            'tts', 'components'))
                        response_parameters['show_for_invoking_user_only'] = \
                            self._parameters.get('show_for_invoking_user_only', show_for_invoking_user_only)
                        
                        yield client.interaction_response_message_create(interaction_event, **response_parameters)
                    return
                
                response_parameters = self._get_response_parameters(('allowed_mentions', 'content', 'embed', 'file',
                    'tts', 'components',))
                response_parameters['show_for_invoking_user_only'] = \
                    self._parameters.get('show_for_invoking_user_only', show_for_invoking_user_only)
                
                yield client.interaction_followup_message_create(interaction_event, **response_parameters)
                return
        
        else:
            if response_state == INTERACTION_EVENT_RESPONSE_STATE_NONE:
                if 'file' in self._parameters:
                    show_for_invoking_user_only = \
                        self._parameters.get('show_for_invoking_user_only', show_for_invoking_user_only)
                    
                    yield client.interaction_response_message_create(interaction_event,
                        show_for_invoking_user_only=show_for_invoking_user_only)
                    
                    response_parameters = self._get_response_parameters(('allowed_mentions', 'content', 'embed', 'file',
                        'tts', 'components'))
                    
                    if 'components' in response_parameters:
                        # Enforce new message.
                        yield client.interaction_followup_message_create(interaction_event, **response_parameters)
                    else:
                        yield client.interaction_response_message_edit(interaction_event, **response_parameters)
                else:
                    response_parameters = self._get_response_parameters(('allowed_mentions', 'content', 'embed', 'tts',
                        'components',))
                    response_parameters['show_for_invoking_user_only'] = \
                        self._parameters.get('show_for_invoking_user_only', show_for_invoking_user_only)
                    
                    yield client.interaction_response_message_create(interaction_event, **response_parameters)
                return
            
            if response_state == INTERACTION_EVENT_RESPONSE_STATE_DEFERRED:
                response_parameters = self._get_response_parameters(('allowed_mentions', 'content', 'embed' 'file',
                    'components',))
                
                if 'components' in response_parameters:
                    # Enforce new message.
                    yield client.interaction_followup_message_create(interaction_event, **response_parameters)
                else:
                    yield client.interaction_response_message_edit(interaction_event, **response_parameters)
                return
            
            if response_state == INTERACTION_EVENT_RESPONSE_STATE_RESPONDED:
                response_parameters = self._get_response_parameters(('allowed_mentions', 'content', 'embed', 'file',
                    'tts', 'components'))
                response_parameters['show_for_invoking_user_only'] = \
                    self._parameters.get('show_for_invoking_user_only', show_for_invoking_user_only)
                
                yield client.interaction_followup_message_create(interaction_event, **response_parameters)
                return
    
    def __repr__(self):
        """Returns the slash response's representation."""
        result = ['<', self.__class__.__name__, ' ']
        if self._force_new_message:
            result.append('(force new message) ')
        
        parameters = self._parameters
        if parameters:
            for key, value in parameters.items():
                result.append(key)
                result.append('=')
                result.append(repr(value))
                result.append(', ')
            
            result[-1] = '>'
        else:
            result.append('>')
        
        return ''.join(result)


def abort(content=..., *, embed=..., file=..., allowed_mentions=..., tts=..., show_for_invoking_user_only=...,):
    """
    Aborts the slash response with sending the passed parameters as a response.
    
    The abortion auto detects `show_for_invoking_user_only` if not given. Not follows the command's preference.
    If only a string `content` is given, `show_for_invoking_user_only` will become `True`, else `False`. The reason of
    becoming `False` at that case is, Discord ignores every other field except string content.
    
    Parameters
    ----------
    content : `str`, ``EmbedBase``, `Any`, Optional
        The message's content if given. If given as `str` or empty string, then no content will be sent, meanwhile
        if any other non `str` or ``EmbedBase`` instance is given, then will be casted to string.
        
        If given as ``EmbedBase`` instance, then is sent as the message's embed.
    embed : ``EmbedBase`` instance or `list` of ``EmbedBase`` instances, Optional (Keyword only)
        The embedded content of the message.
        
        If `embed` and `content` parameters are both given as  ``EmbedBase`` instance, then `TypeError` is raised.
    file : `Any`, Optional (Keyword only)
        A file to send. Check ``Client._create_file_form`` for details.
    allowed_mentions : `None`,  `str`, ``UserBase``, ``Role``, `list` of (`str`, ``UserBase``, ``Role`` )
            , Optional (Keyword only)
        Which user or role can the message ping (or everyone). Check ``Client._parse_allowed_mentions`` for details.
    tts : `bool`, Optional (Keyword only)
        Whether the message is text-to-speech.
    show_for_invoking_user_only : `bool`, Optional (Keyword only)
        Whether the sent message should only be shown to the invoking user.
        
        If given as `True`, only the message's content will be processed by Discord.
    
    Raises
    ------
    InteractionAbortedError
        The exception which aborts the interaction, then yields the response.
    """
    if show_for_invoking_user_only is ...:
        if (embed is not ...):
            show_for_invoking_user_only = False
        elif (file is not ...):
            show_for_invoking_user_only = False
        elif (allowed_mentions is not ...):
            show_for_invoking_user_only = False
        elif (tts is not ...):
            show_for_invoking_user_only = False
        elif (content is ...):
            show_for_invoking_user_only = True
        elif is_only_embed(content):
            show_for_invoking_user_only = False
        else:
            show_for_invoking_user_only = True
    
    
    response = SlashResponse(content, embed=embed, file=file, allowed_mentions=allowed_mentions, tts=tts,
        show_for_invoking_user_only=show_for_invoking_user_only, force_new_message=-1)
    
    raise InteractionAbortedError(response)


class InteractionAbortedError(BaseException):
    """
    An ``InteractionAbortedError`` is raised when a slash command is aborted. This class holds the response to send to
    the client.
    
    Attributes
    ----------
    response : ``SlashResponse``
        The response to send.
    """
    def __init__(self, response):
        """
        Creates a new ``InteractionAbortedError`` instance with the given response.
        
        Parameters
        ----------
        response : ``SlashResponse``
            The response to send.
        """
        self.response = response
        BaseException.__init__(self, response)
    
    def __repr__(self):
        """Returns the exception's representation."""
        return f'{self.__class__.__name__}({self.response!r})'



# -*- coding: utf-8 -*-
import re
from datetime import datetime, timezone
from collections import defaultdict
from http.cookies import Morsel, SimpleCookie
from math import ceil

from .url import URL
from .helpers import is_ip_address
from .event_loop import LOOP_TIME

class CookieJar:
    """
    Implements cookie storage adhering to RFC 6265.
    
    Attributes
    ----------
    cookies : `defaultdict` of (`str`, `http.cookies.SimpleCookie`) items
        Dictionary containing the stored cookies.
    expirations : `dict` of (`tuple` (`str`, `str`), `float`) item
        Cookie expirations. The keys are a domain, name pairs, meanwhile the values are their expirations.
    host_only_cookies : `set` of `tuple` (`str`, `str`)
        Reference of hosts only cookies in tuples as `host-name`, `host` pairs.
    next_expiration : `float`
        The next expiration's monotonic time.
    unsafe : `bool`
        Whether the jar accepts unsafe cookies as well.
    """
    DATE_TOKENS_RE = re.compile(
        '[\x09\x20-\x2F\x3B-\x40\x5B-\x60\x7B-\x7E]*(?P<token>[\x00-\x08\x0A-\x1F\d:a-zA-Z\x7F-\xFF]+)')
    DATE_HMS_TIME_RE = re.compile('(\d{1,2}):(\d{1,2}):(\d{1,2})')
    DATE_DAY_OF_MONTH_RE = re.compile('(\d{1,2})')
    DATE_MONTH_RE = re.compile('(jan)|(feb)|(mar)|(apr)|(may)|(jun)|(jul)|(aug)|(sep)|(oct)|(nov)|(dec)', re.I)
    DATE_YEAR_RE = re.compile('(\d{2,4})')
    MAX_TIME = 2051215261.0 # (2035-01-01)
    
    __slots__ = ('cookies', 'expirations', 'host_only_cookies', 'next_expiration', 'unsafe', )
    
    def __init__(self, unsafe=False):
        """
        Creates a new cookie-jar.
        
        Parameters
        ----------
        unsafe : `bool`
            Whether unsafe cookies are added to the jar as well. Defaults to `False`.
        """
        self.cookies = defaultdict(SimpleCookie)
        self.host_only_cookies = set()
        self.unsafe = unsafe
        self.next_expiration = ceil(LOOP_TIME())
        self.expirations = {}
    
    def __getstate__(self):
        return (list(self.cookies.values()), self.host_only_cookies, self.unsafe)
    
    def __setstate__(self, state):
        cookies, host_only_cookies, unsafe = state
        
        self.cookies = defaultdict(SimpleCookie)
        self.host_only_cookies = host_only_cookies
        self.unsafe = False
        self.next_expiration = 0.0
        self.expirations = {}
        
        for cookie in cookies:
            self.update_cookies(cookie)
        
        self.unsafe = unsafe
        self._do_expiration()
    
    def clear(self):
        """
        Clears the cookie-jar.
        """
        self.cookies.clear()
        self.host_only_cookies.clear()
        self.next_expiration = ceil(LOOP_TIME())
        self.expirations.clear()
        
    def __iter__(self):
        """
        Iterates over the jar's cookies' values.
        
        This method is a generator.
        
        Returns
        -------
        value : `str`
        """
        self._do_expiration()
        for val in self.cookies.values():
            yield from val.values()
    
    def __len__(self):
        """Returns the length of the cookie-jar."""
        length = 0
        for cookie in self.cookies.values():
            length += len(cookie)
        
        return length
    
    def _do_expiration(self):
        """
        Removes the expired cookies from the jar.
        """
        now = LOOP_TIME()
        if self.next_expiration > now:
            return
        
        expirations = self.expirations
        if not expirations:
            return
        
        next_expiration = self.MAX_TIME
        to_del = []
        cookies = self.cookies
        
        for key, when in expirations.items():
            if when < now:
                cookies[key[0]].pop(key[1], None)
                to_del.append(key)
                self.host_only_cookies.discard(key)
            else:
                next_expiration = min(next_expiration, when)
        
        for key in to_del:
            del expirations[key]
        
        self.next_expiration = ceil(next_expiration)
    
    def _expire_cookie(self, when, domain, name):
        """
        Sets the expiration of the cookie jar and of the given `domain`, `name` relation.
        
        Parameters
        ----------
        when : `int`
            Posix timestamp of the cookie's expiration.
        domain : `str`
            The cookie's domain.
        name : `str`
            The cookie's name.
        """
        self.next_expiration = min(self.next_expiration, when)
        self.expirations[(domain, name)] = when
    
    def update_cookies(self, cookies, response_url=URL()):
        """
        Updates the cookies of the.
        
        Parameters
        ----------
        cookies : `http.cookies.SimpleCookie`
            A cookie to update the
        response_url : ``URL``
            Respective response's url.
        """
        hostname = response_url.raw_host
        
        if not self.unsafe and is_ip_address(hostname):
            # Don't accept cookies from IPs
            return
        
        for name, cookie in cookies.items():
            if not isinstance(cookie, Morsel):
                temporary = SimpleCookie()
                temporary[name] = cookie
                cookie = temporary[name]
            
            domain = cookie['domain']
            
            # ignore domains with trailing dots
            if domain.endswith('.'):
                domain = ''
                del cookie['domain']
            
            if not domain and hostname is not None:
                # Set the cookie's domain to the response hostname and set its host-only-flag
                self.host_only_cookies.add((hostname, name))
                domain = cookie['domain'] = hostname
            
            if domain.startswith('.'):
                # Remove leading dot
                domain = domain[1:]
                cookie['domain'] = domain
            
            if hostname and not self._is_domain_match(domain, hostname):
                # Setting cookies for different domains is not allowed
                continue
            
            path = cookie['path']
            if not path or not path.startswith('/'):
                # Set the cookie's path to the response path
                path = response_url.path
                if not path.startswith('/'):
                    path = '/'
                else:
                    # Cut everything from the last slash to the end
                    path = f'/{path[1:path.rfind("/")]}'
                cookie['path'] = path
            
            max_age = cookie['max-age']
            if max_age:
                try:
                    delta_seconds = int(max_age)
                    self._expire_cookie(LOOP_TIME()+delta_seconds, domain, name)
                except ValueError:
                    cookie['max-age'] = ''
            
            else:
                expires = cookie['expires']
                if expires:
                    expire_time = self._parse_date(expires)
                    if expire_time is None:
                        cookie['expires'] = ''
                    else:
                        self._expire_cookie(expire_time.timestamp(), domain, name)
            
            self.cookies[domain][name] = cookie
        
        self._do_expiration()
    
    def filter_cookies(self, request_url=URL()):
        """
        Returns this jar's cookies filtered by their attributes
        
        Parameters
        ----------
        request_url : ``URL``
            The url to filter the cookies by.
        
        Returns
        -------
        filtered : ``SimpleCookie``
            Filtered out cookies.
        """
        self._do_expiration()
        
        filtered = SimpleCookie()
        hostname = request_url.raw_host
        if hostname is None:
            hostname = ''
        
        is_not_secure = (request_url.scheme not in ('https', 'wss'))
        
        for cookie in self:
            name = cookie.key
            domain = cookie['domain']
            
            # Send shared cookies
            if not domain:
                filtered[name] = cookie.value
                continue
            
            if not self.unsafe and is_ip_address(hostname):
                continue
            
            if (domain, name) in self.host_only_cookies:
                if domain != hostname:
                    continue
            
            elif not self._is_domain_match(domain, hostname):
                continue
            
            if not self._is_path_match(request_url.path, cookie['path']):
                continue
            
            if is_not_secure and cookie['secure']:
                continue
            
            # It's critical we use the Morsel so the coded_value (based on cookie version) is preserved
            morsel_value = cookie.get(cookie.key, Morsel())
            morsel_value.set(cookie.key, cookie.value, cookie.coded_value)
            filtered[name] = morsel_value
        
        return filtered
    
    @staticmethod
    def _is_domain_match(domain, hostname):
        """
        Implements domain matching adhering to RFC 6265.
        
        Parameters
        ----------
        domain : `str`
            The domain's name to match.
        hostname : `str`
            The hostname to match.
        
        Returns
        -------
        domain_matching : `bool`
        """
        if hostname == domain:
            return True

        if not hostname.endswith(domain):
            return False

        non_matching = hostname[:-len(domain)]

        if not non_matching.endswith('.'):
            return False

        return not is_ip_address(hostname)

    @staticmethod
    def _is_path_match(request_path, cookie_path):
        """
        Implements path matching adhering to RFC 6265.
        
        Parameters
        ----------
        request_path : `str`
            The request's path.
        cookie_path : `str`
            The cookie's path.
        
        Returns
        -------
        path_matching : `bool`
        """
        if not request_path.startswith('/'):
            request_path = '/'
        
        if request_path == cookie_path:
            return True
        
        if not request_path.startswith(cookie_path):
            return False

        if cookie_path.endswith('/'):
            return True
        
        cookie_path_ln = len(cookie_path)
        if cookie_path_ln >= len(request_path):
            return False
        
        if request_path[cookie_path_ln] == '/':
            return True
        
        return False
    
    @classmethod
    def _parse_date(cls, date_str):
        """
        Implements date string parsing adhering to RFC 6265.
        
        Parameters
        ----------
        date_str : `str`
            Datetime string.
        
        Returns
        -------
        date : `None` or `datetime`
            The parsed date. If no date is given or if the `date_str` is invalid, returns `None`,
        """
        if not date_str:
            return
        
        found_time = found_day = found_month = found_year = False
        
        hour = minute = second = day = month = year = 0
        
        for token_match in cls.DATE_TOKENS_RE.finditer(date_str):
            
            token = token_match.group('token')
            
            if not found_time:
                time_match = cls.DATE_HMS_TIME_RE.match(token)
                if time_match:
                    found_time = True
                    hour, minute, second = time_match.groups()
                    hour = int(hour)
                    minute = int(minute)
                    second = int(second)
                    continue
            
            if not found_day:
                day_match = cls.DATE_DAY_OF_MONTH_RE.match(token)
                if day_match:
                    found_day = True
                    day = int(day_match.group())
                    continue
            
            if not found_month:
                month_match = cls.DATE_MONTH_RE.match(token)
                if month_match:
                    found_month = True
                    month = month_match.lastindex
                    continue
            
            if not found_year:
                year_match = cls.DATE_YEAR_RE.match(token)
                if year_match:
                    found_year = True
                    year = int(year_match.group())
        
        if 70 <= year <= 99:
            year += 1900
        elif 0 <= year <= 69:
            year += 2000
        
        if (not found_day) or (not found_month) or (not found_year) or (not found_time) or (year < 1601) or (day < 1) \
                or (day > 31) or (hour > 23) or (minute > 59) or (second > 59):
            return None
        
        return datetime(year, month, day, hour, minute, second, tzinfo=timezone.utc)



# -*- coding: utf-8 -*-
import socket as module_socket
from http.cookies import SimpleCookie

try:
    import ssl as module_ssl
except ImportError:
    module_ssl = None
    ssl_errors = ()
else:
    ssl_errors = (module_ssl.SSLError, module_ssl.CertificateError)
    
    SSL_CONTEXT_VERIFIED = module_ssl.create_default_context()
    
    SSL_CONTEXT_UNVERIFIED = module_ssl.SSLContext(module_ssl.PROTOCOL_SSLv23)
    SSL_CONTEXT_UNVERIFIED.options |= module_ssl.OP_NO_SSLv2|module_ssl.OP_NO_SSLv3|module_ssl.OP_NO_COMPRESSION
    SSL_CONTEXT_UNVERIFIED.set_default_verify_paths()

from .utils import imultidict
from .futures import shield, Task
from .event_loop import LOOP_TIME

from .headers import HOST, METHOD_GET, AUTHORIZATION, PROXY_AUTHORIZATION, METHOD_CONNECT
from .reqrep import ClientRequest, Fingerprint, SSL_ALLOWED_TYPES
from .exceptions import ProxyError
from .helpers import is_ip_address
from .websocket import ProtocolBase

KEEP_ALIVE_TIMEOUT = 15.0

class Connection:
    """
    Reusable connection used by connectors to remember their details.
    
    Attributes
    ----------
    callbacks : `list` of `callable`
        Callable-s to run when the connection is ``.close``-d, ``.release``-d or ``.detach``-ed. They should accept no
        parameters.
    connector : ``ConnectorBase`` instance
        The respective connector of the connection.
    key : ``ConnectionKey``
        A key which contains information about the host.
    protocol : `None` or ``ReadProtocolBase`` instance
        The connection's actual protocol. Set as `None` if the connection is ``.close``-d, ``.release``-d or
        ``.detach``-ed.
    """
    __slots__ = ('callbacks', 'connector', 'key', 'protocol',)
    def __init__(self, connector, key, protocol):
        """
        Creates a new ``Connection`` instance with the given parameters.
        
        connector : ``ConnectorBase`` instance
            The respective connector of the connection.
        key : `tuple` (`str`, `int`)
            The host and the port to what the connection is connected too.
        protocol : ``ReadProtocolBase`` instance
            The connection's actual protocol.
        """
        self.connector = connector
        self.key = key
        self.protocol = protocol
        self.callbacks = []
    
    def __repr__(self):
        """Returns the representation of the connection."""
        result = [
            '<',
            self.__class__.__name__,
            ' to ',
            repr(self.key),
                ]
        
        if self.closed:
            result.append(' closed')
        
        result.append('>')
        
        return ''.join(result)
    
    def __del__(self):
        """Releases the connector if not yet released."""
        connector = self.connector
        if connector.loop.running:
            protocol = self.protocol
            if (protocol is not None):
                connector.release(self.key, protocol, should_close=True)
    
    @property
    def transport(self):
        """
        Returns the connection's ``.protocol``'s transport if applicable.
        
        Returns
        -------
        transport : `Any`
            Defaults to `None`.
        """
        protocol = self.protocol
        if protocol is None:
            return None
        
        return protocol.transport
    
    @property
    def writer(self):
        """
        Returns the connection's ``.protocol``'s writer if applicable.
        
        Returns
        -------
        writer : `Any`
            Defaults to `None`.
        """
        protocol = self.protocol
        if protocol is None:
            return None
        
        return protocol.writer
    
    def add_callback(self, callback):
        """
        Adds a callback to the connection. If the connection is already closed runs it instantly.
        
        Parameters
        ----------
        callback : `callable`
            Callable to run when the connection is ``.close``-d, ``.release``-d or ``.detach``-ed. Should accept no
            parameters.
        """
        if callback is None:
            return
        
        if not self.closed:
            self.callbacks.append(callback)
            return
        
        try:
            callback()
        except BaseException as err:
            self.connector.loop.render_exc_async(err, [
                'Exception occurred at ',
                repr(self),
                '.add_callback\nAt running ',
                repr(callback),
                '\n',
                    ])
    
    def _run_callbacks(self):
        """
        Runs the added callbacks of the connection.
        """
        callbacks = self.callbacks
        while callbacks:
            callback = callbacks.pop()
            
            try:
                callback()
            except BaseException as err:
                self.connector.loop.render_exc_async(err, [
                    'Exception occurred at ',
                    repr(self),
                    '._run_callbacks\nAt running ',
                    repr(callback),
                    '\n',
                        ])
    
    def close(self):
        """
        Closes the connection by running it's callbacks and releasing with the intent of closing.
        """
        self._run_callbacks()
        
        protocol = self.protocol
        if (protocol is not None):
            self.protocol = None
            self.connector.release(self.key, protocol, should_close=True)
    
    def release(self):
        """
        Closes the connection by running it's callbacks and releasing it.
        """
        self._run_callbacks()
        
        protocol = self.protocol
        if (protocol is not None):
            self.protocol = None
            self.connector.release(self.key, protocol, should_close=protocol.should_close())
    
    def detach(self):
        """
        Detaches the connection from it ``.connector``.
        
        Note, that with detaching, the connection set it's ``.protocol`` as `None`, so you should save an instance of
        it down before.
        """
        self._run_callbacks()
        
        protocol = self.protocol
        if (protocol is not None):
            self.protocol = None
            self.connector.release_acquired_protocols(self.key, protocol)
   
    @property
    def closed(self):
        """
        Returns whether the connection is closed.
        
        Returns
        -------
        closed : `bool`
        """
        protocol = self.protocol
        if protocol is None:
            return True
        
        if protocol.transport is None:
            return True
        
        return False


class ConnectorBase:
    """
    Base connector class.
    
    Attributes
    ----------
    acquired_protocols : `set` of ``ReadProtocolBase``
        Acquired protocols of the connector.
    acquired_protocols_per_host : `dict` of (``ConnectionKey``, `set` of ``ReadProtocolBase``) items
        Acquired protocols for each host.
    alive_protocols_per_host : `dict` of (``ConnectionKey``, `list` of `tuple` (``ReadProtocolBase``, `float`)) items
        Alive, not used protocols for each host. Each element of the values stores when the connection was last used
        as well.
    cleanup_handle : `None` or ``TimerWeakHandle``
        Weak handle, which cleans up the timed out connections of the connector.
    closed : `bool`
        Whether the connector is closed.
    cookies : `http.cookies.SimpleCookie`
        Cookies of the connection.
    force_close : `bool`
        Whether after each request (and between redirects) the connections should be closed.
    loop : ``EventThread``
        The event loop to what the connector is bound to.
    
    Notes
    -----
    Connectors support weakreferencing.
    """
    __slots__ = ('__weakref__', 'acquired_protocols', 'acquired_protocols_per_host', 'alive_protocols_per_host',
        'cleanup_handle', 'closed',  'cookies', 'force_close', 'loop', )
    
    def __new__(cls, loop, force_close=False,):
        """
        Creates a new connector bound to the given loop.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the connector is bound to.
        force_close : `bool`, Optional
            Whether after each request (and between redirects) the connections should be closed. Defaults to `False`.
        """
        self = object.__new__(cls)
        self.loop = loop
        self.closed = False
        self.alive_protocols_per_host = {}
        self.acquired_protocols = set()
        self.acquired_protocols_per_host = {}
        self.force_close = force_close
        self.cookies = SimpleCookie()
        self.cleanup_handle = None
        return self
    
    def __del__(self):
        """
        Closes the connector if not yet closed.
        """
        if (not self.closed) and self.alive_protocols_per_host:
            self.close()
    
    def _cleanup(self):
        """
        Cleans ups the expired connections of the connector.
        """
        handle = self.cleanup_handle
        if (handle is not None):
            # Cancelling the currently running handle does nothing.
            handle.cancel()
            self.cleanup_handle = None
        
        # Cleanup unused transports.
        alive_protocols_per_host = self.alive_protocols_per_host
        if not alive_protocols_per_host:
            return
        
        now = LOOP_TIME()
        to_remove_keys = []
        
        for key, alive_protocols_for_host in alive_protocols_per_host.items():
            for index in reversed(range(len(alive_protocols_for_host))):
                protocol, time = alive_protocols_for_host[index]
                
                if time + KEEP_ALIVE_TIMEOUT < now:
                    continue
                
                del alive_protocols_for_host[index]
                transport = protocol.transport
                if key.is_ssl and (transport is not None):
                    transport.abort()
            
            if not alive_protocols_for_host:
                to_remove_keys.append(key)
        
        for key in to_remove_keys:
            del alive_protocols_per_host[key]
        
        if alive_protocols_per_host:
            self.cleanup_handle = self.loop.call_later_weak(KEEP_ALIVE_TIMEOUT, self._cleanup,)
    
    def close(self):
        """
        Closes the connector and it's connections.
        """
        if self.closed:
            return
        
        self.closed = True
        
        try:
            if not self.loop.running:
                return
            
            for alive_protocols_for_host in self.alive_protocols_per_host.values():
                for protocol, time in alive_protocols_for_host:
                    protocol.close()
            
            for protocol in self.acquired_protocols:
                protocol.close()
        
        finally:
            self.alive_protocols_per_host.clear()
            self.acquired_protocols.clear()
    
    async def connect(self, request):
        """
        Gets connection from connection pool or creates a new one.
        
        This method is a coroutine.
        
        Parameters
        ----------
        request : ``ClientRequest``
            Respective request, which requires a connection.
        
        Returns
        -------
        connection : ``Connection``
            The created connection.
        
        Raises
        ------
        ConnectionError
            Connector closed.
        """
        key = request.connection_key
        
        protocol = self.get_protocol(key)
        if protocol is None:
            protocol = await self.create_connection(request)
            if self.closed:
                protocol.close()
                raise ConnectionError('Connector is closed.')
        
        self.acquired_protocols.add(protocol)
        
        acquired_protocols_per_host = self.acquired_protocols_per_host
        try:
            acquired_protocols_for_host = acquired_protocols_per_host[key]
        except KeyError:
            acquired_protocols_for_host = acquired_protocols_per_host[key] = set()
        
        acquired_protocols_for_host.add(protocol)
        
        return Connection(self, key, protocol)
    
    def get_protocol(self, key):
        """
        Gets a protocol for the given connection key.
        
        Parameters
        ----------
        key : ``ConnectionKey``
            A key which contains information about the host.

        Returns
        -------
        protocol : `None` or ``ReadProtocolBase`` instance
            Protocol connected to the respective host. Defaults to `None` if there is not any.
        """
        try:
            alive_protocols_for_host = self.alive_protocols_per_host[key]
        except KeyError:
            return None
        
        now = LOOP_TIME()
        while alive_protocols_for_host:
            protocol, time = alive_protocols_for_host.pop()
            if (protocol.transport is None):
                continue
            
            if (now - time) > KEEP_ALIVE_TIMEOUT:
                transport = protocol.transport
                protocol.close()
                if key.is_ssl and (transport is not None):
                    transport.abort()
            else:
                if not alive_protocols_for_host:
                    del self.alive_protocols_per_host[key]
                
                return protocol
        
        del self.alive_protocols_per_host[key]
    
    def release_acquired_protocols(self, key, protocol):
        """
        Removes the given acquired protocol from the connector.
        
        Parameters
        ----------
        key : ``ConnectionKey``
            A key which contains information about the host.
        protocol : ``ReadProtocolBase`` instance
            The connected protocol to the respective host.
        """
        if self.closed:
            return
        
        self.acquired_protocols.discard(protocol)
        
        acquired_protocols_per_host = self.acquired_protocols_per_host
        try:
            acquired_protocols_for_host = acquired_protocols_per_host[key]
        except KeyError:
            pass
        else:
            try:
                acquired_protocols_for_host.remove(protocol)
            except KeyError:
                pass
            else:
                if not acquired_protocols_for_host:
                    del acquired_protocols_per_host[key]
        
    def release(self, key, protocol, should_close=False):
        """
        Releases the given protocol from the connector. If the connection should not be closed, not closes it, instead
        stores it for future reuse.
        
        Parameters
        ----------
        key : ``ConnectionKey``
            A key which contains information about the host.
        protocol : ``ReadProtocolBase``
            Protocol of the released connection.
        should_close : `bool`
            Whether the respective connection should be closed.
        """
        if self.closed:
            return
        
        self.release_acquired_protocols(key, protocol)
        
        if should_close or self.force_close or protocol.should_close():
            transport = protocol.transport
            protocol.close()
            if key.is_ssl and (transport is not None):
                transport.abort()
        else:
            try:
                alive_protocols_for_host = self.alive_protocols_per_host[key]
            except KeyError:
                alive_protocols_for_host = self.alive_protocols_per_host[key] = []
            
            alive_protocols_for_host.append((protocol, LOOP_TIME()))
            
            if self.cleanup_handle is None:
                self.cleanup_handle = self.loop.call_later_weak(KEEP_ALIVE_TIMEOUT, self._cleanup,)
    
    async def create_connection(self, request):
        """
        Creates a new connection for the given request.
        
        This method is a coroutine.
        
        Parameters
        ----------
        request : ``ClientRequest``
            Request to create connection for.
        
        Returns
        -------
        protocol : ``ReadProtocolBase`` instance
            The created protocol connected to the respective host.
        
        Raises
        ------
        NotImplementedError
            ``ConnectorBase`` do not implements ``.create_connection``. Subclasses should do it.
        """
        raise NotImplementedError

DNS_CACHE_TIMEOUT = 10.0

class HostInfo:
    """
    Resolved information about a host.
    
    Attributes
    ----------
    hostname : `str`
        The hosts's name.
    host : `str`
        The host's ip address.
    port : `int`
        Port to connect to the host.
    family : `AddressFamily` or `int`
        Address family.
    protocol : `int`
        Protocol type.
    flags : `int`
        Bit-mask for `get_address_info`.
    """
    __slots__ = ('hostname', 'host', 'port', 'family', 'protocol', 'flags', )
    
    def __repr__(self):
        """Returns the host info's representation."""
        return f'<{self.__class__.__name__}, hostname={self.hostname!r}, host={self.host!r}, port={self.port!r}, ' \
               f'family={self.family!r}, protocol={self.protocol!r}, flags={self.flags!r}>'
    
    @classmethod
    def from_ip(cls, host, port, family):
        """
        Creates a host info instance from the given parameters.
        
        Parameters
        ----------
        host : `str`
            The host's ip address.
        port : `int`
            Port to connect to the host.
        family : `AddressFamily` or `int`
            Address family.
        """
        self = object.__new__(cls)
        self.hostname = host
        self.host = host
        self.port = port
        self.family = family
        self.protocol = 0
        self.flags = 0
        
        return self
    
    @classmethod
    def from_address_info(cls, host, address_info):
        """
        Creates a host info instance from the given parameters.
        
        Parameters
        ----------
        host : `str`
            The host's ip address.
        address_info : `tuple` (`AddressFamily` or `int`, `SocketKind` or `int`, `int`, `str`, `tuple` (`str, `int`))
            An address info returned by `get_address_info`.
        """
        self = object.__new__(cls)
        self.hostname = host
        address = address_info[4]
        self.host = address[0]
        self.port = address[1]
        self.family = address_info[0]
        self.protocol = address_info[2]
        self.flags = module_socket.AI_NUMERICHOST
        
        return self
    
    def __eq__(self, other):
        """Returns whether the two host infos are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.hostname != other.hostname:
            return False
        
        if self.host != other.host:
            return False
        
        if self.port != other.port:
            return False
        
        if self.family != other.family:
            return False
        
        if self.protocol != other.protocol:
            return False
        
        if self.flags != other.flags:
            return False
        
        return True
    
    def __hash__(self):
        """Returns the host infos hash value."""
        return hash(self.hostname) ^ hash(self.host) ^ (self.port << 17 ) ^ hash(self.family) ^ hash(self.protocol) ^ \
               hash(self.flags)


class HostInfoCont:
    """
    ``HostInfo`` container, which rotates it's hosts.
    
    Attributes
    ----------
    index : `int`
        Index to determine since the next yielded addresses will start. Used for cycle host infos.
    timestamp : `int`
        Monotonic time determining when the container was created.
    host_infos : `list` of ``HostInfo``
        A list of the contained host information.
    """
    __slots__ = ('host_infos', 'index', 'timestamp')
    
    def __new__(cls, host, address_infos):
        """
        Creates a new host container from the given parameters.
        
        Parameters
        ----------
        host : `str`
            The host's name.
        address_infos : `list` of tuple` \
                (`AddressFamily` or `int`, `SocketKind` or `int`, `int`, `str`, `tuple` (`str, `int`))
            Address infos returned by `get_address_info`.
        """
        self = object.__new__(cls)
        self.index = 0
        self.timestamp = LOOP_TIME()
        self.host_infos = [HostInfo.from_address_info(host, address_info) for address_info in address_infos]
        
        return self
    
    @property
    def expired(self):
        """
        Returns whether the host info container expired.
        
        Returns
        -------
        expired : `bool`
        """
        if (self.timestamp+DNS_CACHE_TIMEOUT) < LOOP_TIME():
            return True
        
        return False
    
    def next_addresses(self):
        """
        Gets the next cycled addresses from the container.
        
        Returns
        -------
        result : `list` of ``HostInfo``
        """
        result = []
        
        host_infos = self.host_infos
        index = self.index
        limit = len(host_infos)
        left = limit
        
        while True:
            host_info = host_infos[index]
            result.append(host_info)
            index += 1
            if index == limit:
                index = 0
            
            left -= 1
            if left:
                continue
            
            break
        
        self.index = index
        return result

    def __repr__(self):
        """Returns the host info container's representation."""
        return f'<{self.__class__.__name__} host_infos={self.host_infos!r}, index={self.index!r}, ' \
            f'timestamp={self.timestamp!r}>'
    
class TCPConnector(ConnectorBase):
    """
    Base connector class.
    
    Attributes
    ----------
    acquired_protocols : `set` of ``ReadProtocolBase``
        Acquired protocols of the connector.
    acquired_protocols_per_host : `dict` of (``ConnectionKey``, `set` of ``ReadProtocolBase``) items
        Acquired protocols for each host.
    alive_protocols_per_host : `dict` of (``ConnectionKey``, `list` of `tuple` (``ReadProtocolBase``, `float`)) items
        Alive, not used protocols for each host. Each element of the values stores when the connection was last used
        as well.
    cleanup_handle : `None` or ``TimerWeakHandle``
        Weak handle, which cleans up the timed out connections of the connector.
    closed : `bool`
        Whether the connector is closed.
    cookies : `http.cookies.SimpleCookie`
        Cookies of the connection.
    force_close : `bool`
        Whether after each request (and between redirects) the connections should be closed.
    loop : ``EventThread``
        The event loop to what the connector is bound to.
    cached_hosts : `dict` of (`tuple` (`str`, `int`), ``HostInfoCont``) items
        Cached resolved host information.
    dns_events : `dict` of (`tuple` (`str`, `int`), ``Task`` of ``.resolve``) items
        Active host info resolving events of the connector.
    family : `AddressFamily` or `int`
        Address family of the created socket if any.
    local_address : `None` or `tuple` of (`None` or  `str`, `None` or `int`)
        Can be given as a `tuple` (`local_host`, `local_port`) to bind created sockets locally.
    ssl : `ssl.SSLContext`, `bool`, ``Fingerprint``, `NoneType`
        Whether and what type of ssl should the connector use.
    
    Notes
    -----
    Connectors support weakreferencing.
    """
    __slots__ = ('cached_hosts', 'dns_events', 'family', 'local_address',  'ssl', )

    def __new__(cls, loop, family=0, ssl=None, local_address=None, force_close=False, ):
        """
        Creates a new ``TCPConnector`` instance with the given parameters.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the connector is bound to.
        family : `AddressFamily` or `int`
            Address family of the created socket if any. Defaults to `0`.
        ssl : `ssl.SSLContext`, `bool`, ``Fingerprint``, `NoneType`, Optional
            Whether and what type of ssl should the connector use. Defaults to `None`.
        local_address : `None` or `tuple` of (`None` or  `str`, `None` or `int`), Optional
            Can be given as a `tuple` (`local_host`, `local_port`) to bind created sockets locally. Defaults to `None`.
        force_close : `bool`, Optional
            Whether after each request (and between redirects) the connections should be closed. Defaults to `False`.
        """
        if not isinstance(ssl, SSL_ALLOWED_TYPES):
            raise TypeError(f'`ssl` should be one of instance of: {SSL_ALLOWED_TYPES!r}, but got `{ssl!r}` instead.')
        
        self = ConnectorBase.__new__(cls, loop, force_close,)
        
        self.ssl = ssl
        self.cached_hosts = {}
        self.dns_events = {}
        self.family = family
        self.local_address = local_address
        
        return self
    
    def close(self):
        """
        Closes the connector, it's dns lookup events and it's connections.
        """
        for event in self.dns_events.values():
            event.cancel()
        
        ConnectorBase.close(self)
    
    async def resolve(self, host, port):
        """
        Resolves a host and returns it's result.
        
        This method is a coroutine.
        
        Parameters
        ----------
        host : `None` or `str`
            To what network interfaces should the connection be bound.
        port : `None` or `int`
            The port of the `host`.
        
        Returns
        -------
        result : ``HostInfoCont`` or ``BaseException``
            A host info container containing the resolved addresses or the cached exception.
        """
        try:
            infos = await self.loop.get_address_info(host, port, type=module_socket.SOCK_STREAM, family=self.family)
        except BaseException as err:
            return err
        
        return HostInfoCont(host, infos,)
    
    async def resolver_task(self, key,):
        """
        Resolves addresses for the given `host` - `port` key.
        
        If there is a ``.resolve`` task for the given `key`, then awaits on that's result instead.
        
        This method is a coroutine.
        
        Parameters
        ----------
        key : `tuple`, (`None` or `str`, `None` or `int`)
            A tuple containing a `host`, `port` par to resolve.
        
        Returns
        -------
        result : ``HostInfoCont`` or ``BaseException``
            A host info container containing the resolved addresses or the cached exception.
        """
        try:
            event = self.dns_events[key]
        except KeyError:
            event = Task(self.resolve(*key), self.loop)
            self.dns_events[key] = event
            host_info = await event
            if type(host_info) is HostInfoCont:
                self.cached_hosts[key] = host_info
            del self.dns_events[key]
        else:
            host_info = await event
        
        return host_info
    
    async def resolve_host_iterator(self, request):
        """
        Yields host infos to which the request can connect to.
        
        This method is a coroutine generator.
        
        Parameters
        ----------
        request : ``ClientRequest``
            Respective request, which requires a connection.
        
        Yields
        ------
        host_info : ``HostInfo``
            Resolved host information.
        
        Raises
        ------
        ConnectionError
            No hosts were resolved.
        BaseException
            Any other cached exception.
        """
        host = request.url.raw_host
        port = request.port
        
        if is_ip_address(host):
            yield HostInfo.from_ip(host, port, self.family)
            return
        
        key = (host, port)
        try:
            host_infos = self.cached_hosts[key]
        except KeyError:
            pass
        else:
            expired = host_infos.expired
            if expired:
                task = shield(self.resolver_task(key), self.loop)
            
            address = host_infos.next_addresses()
            for host_info in address:
                yield host_info
            
            if expired:
                host_infos = await task
                if (type(host_infos) is not HostInfoCont):
                    if isinstance(host_infos, OSError):
                        raise ConnectionError(request.connection_key, host_infos) from host_infos
                    else:
                        raise host_infos
                
                for host_info in host_infos.next_addresses():
                    if host_info in address:
                        continue
                    
                    yield host_info
            
            return
        
        task = shield(self.resolver_task(key), self.loop)
        
        host_infos = await task
        if (type(host_infos) is not HostInfoCont):
            if isinstance(host_infos, OSError):
                raise ConnectionError(request.connection_key, host_infos) from host_infos
            else:
                raise host_infos
        
        for host_info in host_infos.next_addresses():
            yield host_info
    
    async def create_connection(self, request):
        """
        Creates connection for the request and returns the created protocol.
        
        This method is a coroutine.
        
        Parameters
        ----------
        request : ``ClientRequest``
            Respective request.
        
        Returns
        -------
        protocol : ``ReadProtocolBase``
        
        Raises
        ------
        ssl.SSLError
        ssl.CertificateError
        OSError
        RuntimeError
            If transport does not expose socket instance.
        """
        if request.proxy_url is None:
            coro = self.create_direct_connection(request)
        else:
            coro = self.create_proxy_connection(request)
        
        _, protocol = await coro
        return protocol
    
    def get_ssl_context(self, request):
        """
        Gets ssl context for the respective request.
        
        Parameters
        ----------
        request : ``ClientRequest``
            Respective request.
        
        Returns
        -------
        ssl_context : `None` or `ssl.SSLContext`
            Returns `None` if the request is not `ssl`.
        """
        if not request.is_ssl():
            return None
        
        if module_ssl is None:
            raise RuntimeError('SSL is not supported.')
        
        ssl_context = request.ssl
        if (ssl_context is not None):
            if isinstance(ssl_context, module_ssl.SSLContext):
                return ssl_context
            
            return SSL_CONTEXT_UNVERIFIED # not verified or fingerprinted
        
        ssl_context = self.ssl
        if ssl_context is None:
            return SSL_CONTEXT_VERIFIED
        
        if isinstance(ssl_context, module_ssl.SSLContext):
            return ssl_context
        
        return SSL_CONTEXT_UNVERIFIED
    
    def get_fingerprint(self, request):
        """
        Gets fingerprint for the respective request.
        
        Parameters
        ----------
        request : ``ClientRequest``
            Respective request.
        
        Returns
        -------
        fingerprint : `None` or ``Fingerprint``
        """
        maybe_fingerprint = request.ssl
        if isinstance(maybe_fingerprint, Fingerprint):
            return maybe_fingerprint
        
        maybe_fingerprint = self.ssl
        if isinstance(maybe_fingerprint, Fingerprint):
            return maybe_fingerprint
    
    async def create_direct_connection(self, request):
        """
        Creates a direct connection for the given request and returns the created protocol.
        
        This method is a coroutine.
        
        Parameters
        ----------
        request : ``ClientRequest``
            Respective request.
        
        Returns
        -------
        protocol : ``ReadProtocolBase``
        
        Raises
        ------
        ssl.SSLError
        ssl.CertificateError
        OSError
        """
        ssl_context = self.get_ssl_context(request)
        fingerprint = self.get_fingerprint(request)
        
        last_error = None
        
        async for host_info in self.resolve_host_iterator(request):
            try:
                transport, protocol = await self.loop.create_connection(ProtocolBase(self.loop),
                    host_info.host, host_info.port,
                    ssl = ssl_context,
                    family = host_info.family,
                    protocol = host_info.protocol,
                    flags = host_info.flags,
                    local_address = self.local_address,
                    server_hostname = (host_info.hostname if ssl_context else None),
                        )
            except ssl_errors as err:
                err.key = request.connection_key
                raise
            except OSError as err:
                last_error = OSError(request.connection_key, err)
                last_error.__cause__ = err
                continue
            
            if request.is_ssl() and fingerprint:
                try:
                    fingerprint.check(transport)
                except ValueError as err:
                    transport.close()
                    transport.abort()
                    last_error = err
                    continue
            
            return transport, protocol
        
        raise last_error
    
    
    async def create_proxy_connection(self, request):
        """
        Creates a proxy connection for the given request and returns the created protocol.
        
        This method is a coroutine.
        
        Parameters
        ----------
        request : ``ClientRequest``
            Respective request.
        
        Returns
        -------
        protocol : ``ReadProtocolBase``
        
        Raises
        ------
        ssl.SSLError
        ssl.CertificateError
        OSError
        RuntimeError
            If transport does not expose socket instance.
        """
        headers = imultidict()
        
        headers[HOST] = request.headers[HOST]
        
        proxy_request = ClientRequest(self.loop, METHOD_GET, request.proxy_url, headers, None, None, None,
            request.proxy_auth, None, None, request.ssl)
        
        # create connection to proxy server
        transport, protocol = await self.create_direct_connection(proxy_request)
        
        # Many HTTP proxies has buggy keep-alive support. Let's not reuse connection but close it after processing
        # every response.
        protocol.close()
        
        auth = proxy_request.headers.pop(AUTHORIZATION, None)
        if auth is not None:
            if not request.is_ssl():
                request.headers[PROXY_AUTHORIZATION] = auth
            else:
                proxy_request.headers[PROXY_AUTHORIZATION] = auth
        
        if request.is_ssl():
            ssl_context = self.get_ssl_context(request)
            proxy_request.method = METHOD_CONNECT
            proxy_request.url = request.url
            connection = Connection(self, request.connection_key, protocol)
            
            response = await proxy_request.send(connection)
            try:
                connection.protocol = None
                try:
                    if response.status != 200:
                        raise ProxyError(response.status, response.reason, response.headers)
                    raw_socket = transport.get_extra_info('socket', None)
                    if raw_socket is None:
                        raise RuntimeError('Transport does not expose socket instance.')
                    # Duplicate the socket, so now we can close proxy transport
                    raw_socket = raw_socket.dup()
                finally:
                    transport.close()
                
                try:
                    transport, protocol = await self.loop.create_connection(ProtocolBase(self.loop), ssl=ssl_context,
                        socket=raw_socket, server_hostname=request.host,)
                except ssl_errors as err:
                    err.key = request.connection_key
                    raise
                except OSError as err:
                    raise OSError(request.connection_key, err) from err
            
            finally:
                response.close()
        
        return transport, protocol



# -*- coding: utf-8 -*-
__all__ = ('EXTENSION_LOADER', 'EXTENSIONS', 'ExtensionError', 'ExtensionLoader', )

import sys
from io import StringIO
from importlib.util import find_spec, module_from_spec
from importlib import reload as reload_module
from threading import current_thread

from ...backend.event_loop import EventThread
from ...backend.utils import alchemy_incendiary, HybridValueDictionary, WeakValueDictionary
from ...backend.futures import is_coroutine_function as is_coro, Task
from ...backend.analyzer import CallableAnalyzer

from ...discord.client_core import KOKORO

from .snapshot import take_snapshot, calculate_snapshot_difference, revert_snapshot

EXTENSIONS = WeakValueDictionary()

EXTENSION_STATE_UNDEFINED = 0
EXTENSION_STATE_LOADED = 1
EXTENSION_STATE_UNLOADED = 2

class ExtensionError(Exception):
    """
    An exception raised by the ``ExtensionLoader``, if loading, reloading or unloading an extension fails with any
    reason.
    
    Attributes
    ----------
    _message : `str` or `list` of `str`
        The error's message.
    """
    
    def __init__(self, message):
        """
        Creates a new extension error.
        
        Parameters
        ----------
        message : `str` or `list` of `str`
            The error's message.
        """
        self._message = message
    
    @property
    def message(self):
        """
        Returns the extension error's message.
        
        If the extension error contains more message,s connects them.
        
        Returns
        -------
        message : ``Message``
        """
        message = self._message
        if type(message) is str:
            return message
        
        return '\n\n'.join(message)
    
    @property
    def messages(self):
        """
        Returns a list containing the exception error's messages.
        
        Returns
        -------
        messages : `list` of `str`
        """
        message = self._message
        if type(message) is str:
           return [message]
        
        return message
    
    def __len__(self):
        """Returns the amount of messages, what the extension error contains."""
        message = self._message
        if type(message) is str:
            return 1
        
        return len(message)
    
    def __repr__(self):
        """Returns the exception error's representation."""
        return f'{self.__class__.__name__} ({len(self)}):\n{self.message}\n'
    
    __str__ = __repr__

def _validate_entry_or_exit(point):
    """
    Validates the given entry or exit point, returning `True`, if they passed.
    
    Parameters
    ----------
    point : `None`, `str` or `callable`
        The point to validate.
    
    Raises
    ------
    TypeError
        If `point` was given as `callable`, but accepts less or more positional arguments, as would be given.
    """
    if point is None:
        return True
    
    if isinstance(point, str):
        return True
    
    if callable(point):
        analyzer = CallableAnalyzer(point)
        min_, max_ = analyzer.get_non_reserved_positional_argument_range()
        if min_ > 1:
            raise TypeError(f'`{point!r}` excepts at least `{min_!r}` non reserved arguments, meanwhile the event '
                'expects to pass `1`.')
        
        if min_ == 1:
            return True
        
        #min<expected
        if max_ >= 1:
            return True
        
        if analyzer.accepts_args():
            return True
        
        raise TypeError(f'`{point!r}` expects maximum `{max_!r}` non reserved arguments  meanwhile the event expects '
            'to pass `1`.')
    
    return False


class Extension:
    """
    Represents an extension.
    
    Attributes
    ----------
    _added_variable_names : `list of `str`
        A list of the added variables' names to the module.
    _default_variables : `None` or `HybridValueDictionary` of (`str`, `Any`) items
        An optionally weak value dictionary to store objects for assigning them to modules before loading them.
        If it would be set as empty, then it is set as `None` instead.
    _entry_point : `None`, `str`, or `callable`
        Internal slot used by the ``.entry_point`` property.
    _exit_point : `None`, `str`, or `callable`
        Internal slot used by the ``.exit_point`` property.
    _extend_default_variables : `bool`
        Internal slot used by the ``.extend_default_variables`` property.
    _lib : `None` or module`
        The extension's module. Set as `module` object if it the extension was already loaded.
    _locked : `bool`
        The internal slot used for the ``.locked`` property.
    _snapshot_difference : `None` or `list` of `tuple` (``Client``, `list` of `tuple` (`str`, `Any`))
        Snapshot difference if applicable. Defaults to `None`.
    _spec : `ModuleSpec`
        The module specification for the extension's module's import system related state.
    _state : `int`
        The state of the extension. Can be:
        +---------------------------+-------+
        | Respective name           | Value |
        +===========================+=======+
        | EXTENSION_STATE_UNDEFINED | 0     |
        +---------------------------+-------+
        | EXTENSION_STATE_LOADED    | 1     |
        +---------------------------+-------+
        | EXTENSION_STATE_UNLOADED  | 2     |
        +---------------------------+-------+
    _take_snapshot : `bool`
        Whether snapshot difference should be taken.
    """
    __slots__ = ('__weakref__', '_added_variable_names', '_default_variables', '_entry_point', '_exit_point',
        '_extend_default_variables', '_lib', '_locked', '_snapshot_difference', '_spec', '_state',
        '_take_snapshot', )
    
    def __new__(cls, name, entry_point, exit_point, extend_default_variables, locked, take_snapshot_difference,
            default_variables, ):
        """
        Creates an extension with the given parameters. If an extension already exists with the given name, returns
        that.
        
        Parameters
        ----------
        name : `str`
            The extension's name (or import path).
        entry_point : `None`, `str` or `callable`
            The entry point of the extension.
        exit_point : `None`, `str` or `callable`
            The exit point of the extension.
        extend_default_variables : `bool`
            Whether the extension should use the loader's default variables or just it's own's.
        locked : `bool`
            Whether the extension should be picked up by the `{}_all` methods of the extension loader.
        take_snapshot_difference: `bool`
            Whether snapshots should be taken before and after loading an extension, and when the extension is unloaded,
            the snapshot difference should be reverted.
        default_variables : `None` or `HybridValueDictionary` of (`str`, `Any`) items
            An optionally weak value dictionary to store objects for assigning them to modules before loading them.
            If would be empty, is set as `None` instead.
        
        Returns
        -------
        self : ``Extension``
        
        Raises
        ------
        ModuleNotFoundError
            If the extension was not found.
        """
        try:
            return EXTENSIONS[name]
        except KeyError:
            pass
        
        spec = find_spec(name)
        if spec is None:
            raise ModuleNotFoundError(name)
        
        self = object.__new__(cls)
        self._state = EXTENSION_STATE_UNDEFINED
        self._spec = spec
        self._lib = None
        self._entry_point = entry_point
        self._exit_point = exit_point
        self._extend_default_variables = extend_default_variables
        self._locked = locked
        self._default_variables = default_variables
        self._added_variable_names = []
        self._take_snapshot = take_snapshot_difference
        self._snapshot_difference = None
        EXTENSIONS[name] = self
        
        return self
    
    def __hash__(self):
        """Returns the extension's ``._spec``'s `.origin`'s hash."""
        return hash(self._spec.origin)
    
    def __repr__(self):
        """Returns the extension's representation."""
        result = []
        result.append('<')
        result.append(self.__class__.__name__)
        result.append(' name=')
        result.append(repr(self._spec.name))
        
        state = self._state
        result.append(', state=')
        result.append(repr(state))
        result.append(' (')
        result.append(('undefined', 'loaded', 'unloaded')[state])
        result.append(')')
        
        if self._locked:
            result.append(', locked=True')
        
        default_variables = self._default_variables
        if self._extend_default_variables:
            if (default_variables is not None):
                result.append(' extends loader\'s defaults with: ')
                result.append(repr(default_variables))
        else:
            if default_variables is None:
                result.append(' clears loader\'s defaults')
            else:
                result.append(' clears loader\'s defaults and uses: ')
                result.append(repr(default_variables))
        
        result.append('>')
        
        return ''.join(result)
    
    def add_default_variables(self, **variables):
        """
        Adds default variables to the extension.
        
        Parameters
        ----------
        **variables : Keyword Arguments
            Variables to assigned to the extension's module before it is loaded.
        
        Raises
        ------
        ValueError
             If a variable name is would be used, what is `module` attribute.
        """
        if not variables:
            return
        
        default_variables = self._default_variables
        if default_variables is None:
            default_variables = HybridValueDictionary()
            self._default_variables = default_variables
        
        for key, value in variables.items():
            if key in PROTECTED_NAMES:
                raise ValueError(f'The passed {key!r} is a protected variable name of module type.')
            default_variables[key] = value
    
    def remove_default_variables(self, *names):
        """
        Removes the mentioned default variables of the extension.
        
        If a variable with a specified name is not found, no error is raised.
        
        Parameters
        ----------
        *names : `str`
            Default variable names.
        """
        default_variables = self._default_variables
        if default_variables is None:
            return
        
        for name in names:
            try:
                del default_variables[name]
            except KeyError:
                pass
        
        if default_variables:
            return
        
        self._default_variables = None
    
    def clear_default_variables(self):
        """
        Removes all the default variables of the extension.
        """
        self._default_variables = None
    
    @property
    def entry_point(self):
        """
        Get-set-del descriptor for modifying the extension's entry point.
        
        Accepts and returns `None`, `str` or a `callable`. If invalid type is given, raises `TypeError`.
        """
        return self._entry_point
    
    @entry_point.setter
    def entry_point(self, entry_point):
        if not _validate_entry_or_exit(entry_point):
            raise TypeError(f'`{self.__class__.__name__}.entry_point` expected `None`, `str` or a `callable`, got '
                f'{entry_point.__class__.__name__}.')
        
        self._entry_point = entry_point
    
    @entry_point.deleter
    def entry_point(self):
        self._entry_point = None
    
    
    @property
    def exit_point(self):
        """
        Get-set-del descriptor for modifying the extension's exit point.
        
        Accepts and returns `None`, `str` or a `callable`. If invalid type is given, raises `TypeError`.
        """
        return self._exit_point
    
    @exit_point.setter
    def exit_point(self, exit_point):
        if not _validate_entry_or_exit(exit_point):
            raise TypeError(f'`{self.__class__.__name__}.exit_point` expected `None`, `str` or a `callable`, got '
                f'{exit_point.__class__.__name__}.')
        
        self._exit_point = exit_point
    
    @exit_point.deleter
    def exit_point(self):
        self._exit_point = None
    
    
    @property
    def extend_default_variables(self):
        """
        Get-set descriptor to define whether the extension uses the loader's default variables or just it's own's.
        
        Accepts and returns `bool`.
        """
        return self._extend_default_variables
    
    @extend_default_variables.setter
    def extend_default_variables(self, extend_default_variables):
        extend_default_variables_type = extend_default_variables.__class__
        if extend_default_variables_type is bool:
            pass
        elif issubclass(extend_default_variables_type, int):
            extend_default_variables = bool(extend_default_variables)
        else:
            raise TypeError(f'`extend_default_variables` should have been passed as `bool`, got '
                f'{extend_default_variables_type.__name__}.')
        
        self._extend_default_variables = extend_default_variables
    
    @property
    def locked(self):
        """
        Get-set property to define whether the extension should be picked up by the `{}_all` methods of the
        extension loader.
        
        Accepts and returns `bool`.
        """
        return self._locked
    
    @locked.setter
    def locked(self, locked):
        locked_type = locked.__class__
        if locked_type is bool:
            pass
        elif issubclass(locked_type, int):
            locked = bool(locked)
        else:
            raise TypeError(f'`locked` should have been passed as `bool`, got: {locked_type.__name__}.')
        
        self._locked = locked
    
    
    def _load(self):
        """
        Loads the module and returns it. If it is already loaded returns `None`.
        
        Returns
        -------
        lib : `None` or `lib`
        """
        state = self._state
        if state == EXTENSION_STATE_UNDEFINED:
            
            spec = self._spec
            lib = sys.modules.get(spec.name, None)
            if lib is None:
                # lib is not imported yet, nice
                lib = module_from_spec(spec)
                
                added_variable_names = self._added_variable_names
                if self._extend_default_variables:
                    for name, value in EXTENSION_LOADER._default_variables.items():
                        setattr(lib, name, value)
                        added_variable_names.append(name)
                
                default_variables = self._default_variables
                if (default_variables is not None):
                    for name, value in default_variables.items():
                        setattr(lib, name, value)
                        added_variable_names.append(name)
                
                if self._take_snapshot:
                    snapshot_old = take_snapshot()
                
                spec.loader.exec_module(lib)
                sys.modules[spec.name] = lib
            
                if self._take_snapshot:
                    snapshot_new = take_snapshot()
                    
                    self._snapshot_difference = calculate_snapshot_difference(snapshot_old, snapshot_new)
            
            self._lib = lib
            
            self._state = EXTENSION_STATE_LOADED
            return lib
        
        if state == EXTENSION_STATE_LOADED:
            # return None -> already loaded
            return None
        
        if state == EXTENSION_STATE_UNLOADED:
            # reload
            lib = self._lib
            
            added_variable_names = self._added_variable_names
            if self._extend_default_variables:
                for name, value in EXTENSION_LOADER._default_variables.items():
                    setattr(lib, name, value)
                    added_variable_names.append(name)
            
            default_variables = self._default_variables
            if (default_variables is not None):
                for name, value in default_variables.items():
                    setattr(lib, name, value)
                    added_variable_names.append(name)
            
            if self._take_snapshot:
                snapshot_old = take_snapshot()
            
            reload_module(lib)
            
            if self._take_snapshot:
                snapshot_new = take_snapshot()
                
                self._snapshot_difference = calculate_snapshot_difference(snapshot_old, snapshot_new)
            
            self._state = EXTENSION_STATE_LOADED
            return lib
        
        # no more cases
    
    def _unload(self):
        """
        Unloads the module and returns it. If it is already unloaded returns `None`.
        
        Returns
        -------
        lib : `None` or `lib`
        """
        state = self._state
        if state == EXTENSION_STATE_UNDEFINED:
            return None
        
        if state == EXTENSION_STATE_LOADED:
            
            snapshot_difference = self._snapshot_difference
            if (snapshot_difference is not None):
                self._snapshot_difference = None
                revert_snapshot(snapshot_difference)
            
            self._state = EXTENSION_STATE_UNLOADED
            return self._lib
        
        if state == EXTENSION_STATE_UNLOADED:
            return None
        
        # no more cases
    
    def _unassign_variables(self):
        """
        Unassigns the assigned variables to the respective module and clears ``._added_variable_names``.
        """
        lib = self._lib
        if lib is None:
            return
        
        added_variable_names = self._added_variable_names
        for name in added_variable_names:
            try:
                delattr(lib, name)
            except AttributeError:
                pass
        added_variable_names.clear()
    
    @property
    def name(self):
        """
        Returns the extension's name.
        
        Returns
        -------
        name : `str`
        """
        return self._spec.name
    
    def _unlink(self):
        """
        Removes the extension's module from the loaded ones.
        
        Should not be called on loaded extensions.
        """
        state = self._state
        if state == EXTENSION_STATE_UNDEFINED:
            return
        
        if state == EXTENSION_STATE_LOADED:
            self._state = EXTENSION_STATE_UNLOADED
            lib = self._lib
            
            if self._extend_default_variables:
                for name in EXTENSION_LOADER._default_variables:
                    try:
                        delattr(lib, name)
                    except AttributeError:
                        pass
            
            default_variables = self._default_variables
            if (default_variables is not None):
                for name in default_variables:
                    delattr(lib, name)
            
            del sys.modules[self._spec.name]
            return
        
        if state == EXTENSION_STATE_UNLOADED:
            del sys.modules[self._spec.name]
            return
        
        # no more cases

PROTECTED_NAMES = {
    '__class__',
    '__delattr__',
    '__dict__',
    '__dir__',
    '__doc__',
    '__eq__',
    '__format__',
    '__ge__',
    '__getattribute__',
    '__gt__',
    '__hash__',
    '__init__',
    '__init_subclass__',
    '__le__',
    '__lt__',
    '__module__',
    '__ne__',
    '__new__',
    '__reduce__',
    '__reduce_ex__',
    '__repr__',
    '__setattr__',
    '__sizeof__',
    '__str__',
    '__subclasshook__',
    '__weakref__',
    '_cached',
    '_set_fileattr',
    'cached',
    'has_location',
    'loader',
    'loader_state',
    'name',
    'origin',
    'parent',
    'submodule_search_locations',
        }

class ExtensionLoader:
    """
    There are some cases when you probably want to change some functional part of your client in runtime. Load,
    unload or reload code. Hata provides an easy to use (that's what she said) solution to solve this issue.
    
    It is called extension loader is an extension of hata. It is separated from `commands` extension, but it does not
    mean they do not go well together. But what more, extension loader was made to complement it.
    
    Usage
    -----
    The ``ExtensionLoader`` class is instanced already as `EXTENSION_LOADER` and that can be imported as well from
    `extension_loader.py`. Instancing the class again will yield the same object.
    
    Because hata can have more clients, we needed a special extension loader what can handle using more clients at any
    file, so the choice ended up on a  really interesting idea: assign variables to a module before it is (re)loaded.
    
    To show this is not blackmagic, here is an example:
    
    **main.py**
    
    ```py
    from hata.ext.extension_loader import EXTENSION_LOADER
    from datetime import datetime
    
    cake = 'cake'
    now = datetime.now()
    
    EXTENSION_LOADER.add_default_variables(cake=cake, now=now)
    EXTENSION_LOADER.load_extension('extension')
    ```
    
    **extension.py**
    
    ```py
    print(cake, now)
    ```
    
    Make sure you start `main.py` with interactive mode. If you never did it, just use the `-i` option like:
    
    ```
    $ python3 -i main.py
    ```
    
    Or on windows:
    
    ```
    $ python -i main.py
    ```
    
    After you ran `main.py` you should see the following (the date should be different tho):
    
    ```
    cake 2020-03-14 09:40:41.587673
    ```
    
    Because now we have the interpreter, you can change the variables.
    
    ```py
    >>> EXTENSION_LOADER.add_default_variables(cake='cakes taste good, and now is:')
    >>> EXTENSION_LOADER.reload_all()
    ```
    
    And a different text is printed out:
    
    ```
    cakes taste good, and now is: 2020-03-14 09:40:41.587673
    ```
    
    Now lets edit `extension.py`.
    
    ```py
    cake = cake.split()
    print(*cake, now, sep='\\n')
    ```
    
    And reload the extension:
    
    ```py
    >>> EXTENSION_LOADER.reload_all()
    ```
    
    The printed text really changed again:
    
    ```
    cakes
    taste
    good,
    and
    now
    is:
    2020-03-14 09:40:41.587673
    ```
    
    If you remove default variables and the extension file still uses them, you get an ``ExtensionError``:
    
    ```py
    >>> EXTENSION_LOADER.remove_default_variables('cake')
    >>> EXTENSION_LOADER.reload_all()
    ```
    
    ```
    Traceback (most recent call last):
      File "<pyshell#13>", line 1, in <module>
        EXTENSION_LOADER.reload_all()
      File ".../hata/backend/extension_loader.py", line 652, in reload_all
        task.sync_wrap().wait()
      File ".../hata/backend/futures.py", line 823, in wait
        return self.result()
      File ".../hata/backend/futures.py", line 723, in result
        raise exception
      File ".../hata/backend/futures.py", line 1602, in _step
        result=coro.throw(exception)
      File ".../hata/ext/extension_loader/extension_loader.py", line 670, in _reload_all
        raise ExtensionError(error_messages) from None
    hata.ext.extension_loader.ExtensionError: ExtensionError (1):
    Exception occurred meanwhile loading an extension: `extension`.
    
    Traceback (most recent call last):
      File ".../hata/ext/extension_loader/extension_loader.py", line 675, in _load_extension
        lib = await KOKORO.run_in_executor(extension.load)
      File ".../hata/extension_loader.py", line 270, in load
        reload_module(lib)
      File ".../importlib/__init__.py", line 169, in reload
        _bootstrap._exec(spec, module)
      File "<frozen importlib._bootstrap>", line 630, in _exec
      File "<frozen importlib._bootstrap_external>", line 728, in exec_module
      File "<frozen importlib._bootstrap>", line 219, in _call_with_frames_removed
      File ".../extension.py", line 1, in <module>
        cake = cake.split()
    NameError("name 'cake' is not defined")
    ```
    
    Adding Extensions
    -----------------
    Extensions can be added with the `.add` method.
    
    ```py
    EXTENSION_LOADER.add('cute_commands')
    ```
    
    Or more extension can be added as well by passing an iterable:
    
    ```py
    EXTENSION_LOADER.add(['cute_commands', 'nice_commands'])
    ```
    
    If an extension's file is not found, then `.add` will raise  `ModuleNotFoundError`. If the passed argument is not
    `str` instance or not `iterable` of `str`, `TypeError` is raised.
    
    Loading
    -------
    Extensions can be loaded by their name:
    
    ```py
    EXTENSION_LOADER.load('cute_commands')
    ```
    
    All extension can be loaded by using:
    
    ```py
    EXTENSION_LOADER.load_all()
    ```
    
    Or extension can be added and loaded at the same time as well:
    
    ```py
    EXTENSION_LOADER.load_extension('cute_commands')
    ```
    
    `.load_extension` method supports all the keyword arguments as `.add`.
    
    ##### Passing variables to extensions
    
    You can pass variables to extensions with the `.add_default_variables` method:
    
    ```py
    EXTENSION_LOADER.add_default_variables(cake=cake, now=now)
    ```
    
    Adding or removing variables wont change the already loaded extensions' state, those needs to be reloaded to see
    them.
    
    Or pass variables to just specific extensions:
    
    ```py
    EXTENSION_LOADER.add('cute_commands', cake=cake, now=now)
    ```
    
    You can specify if the extension should use just it's own variables and ignore the default ones too:
    
    ```py
    EXTENSION_LOADER.add('cute_commands', extend_default_variables=False, cake=cake, now=now)
    ```
    
    Every variable added is stored in an optionally weak value dictionary, but you are able remove the added variables as well:
    
    ```py
    EXTENSION_LOADER.remove_default_variables('cake', 'now')
    ```
    
    The extensions can be accessed by their name as well, then you can use their properties to modify them.
    
    ```py
    EXTENSION_LOADER.extensions['cute_commands'].remove_default_variables('cake')
    ```
    
    Unloading And Exit Point
    ------------------------
    You can unload extension on the same way as loading them.
    
    ```py
    EXTENSION_LOADER.unload('cute_commands')
    ```
    
    Or unload all:
    
    ```py
    EXTENSION_LOADER.unload_all()
    ```
    
    When unloading an extension, the extension loader will search a function at the extension, what we call
    `exit_point` and will run it. By default it looks for a variable named `teardown`. `exit_point` acts on the same
    way as the `entry_point`, so it can be modified for looking for other name to defining and passing a callable
    (can be async again).
    
    Can be set almost on the same way as well:
    
    ```py
    EXTENSION_LOADER.default_exit_point = 'exit'
    ```
    
    Or:
    
    ```py
    EXTENSION_LOADER.add('cute_commands', exit_point='exit')
    ```
    
    There are also methods for reloading: `.reload(name)` and `.reload_all()`
    
    Removing Extensions
    -------------------
    
    Unloaded extensions can be removed from the extension loader by using the `.remove` method:
    
    ```py
    EXTENSION_LOADER.remove('cute_commands')
    ```
    
    Or more extension with an iterable:
    
    ```py
    EXTENSION_LOADER.remove(['cute_commands', 'nice_commands'])
    ```
    
    Removing loaded extension will yield `RuntimeError`.
    
    Threading Model
    ---------------
    When you call the different methods of the extension loader, they ll run on the ``Clients`` thread, what is named
    `KOKORO` internally.
    
    These methods are:
    - ``.load_extension``
    - ``.load``
    - ``.load_all``
    - ``.unload``
    - ``.unload_all``
    - ``.reload``
    - ``.reload_all``
    
    Meanwhile loading and executing the extension's code the thread is switched to an executor, so blocking tasks can
    be executed easily. The exception under this rule are the `entry_point`-s and the `exit_point`-s, which always run
    on the same thread as the clients, that is why they can be async as well.
    
    These methods also act differently depending from which thread they were called from. Whenever they are called from
    the client's thread, a ``Task`` is returned what can be `awaited`. If called from other ``EventThread``, then the task
    is async_wrapped and that is returned. When calling from any other thread (like the main thread for example), the
    task is sync_wrapped and the thread is blocked till the extension's loading is finished.
    
    Attributes
    ----------
    _default_entry_point : `None`, `str` or `callable`
        Internal slot for the ``.default_entry_point`` property.
    _default_exit_point : `None`, `str` or `callable`
        Internal slot for the ``.default_exit_point`` property.
    _default_variables : `None` or `HybridValueDictionary` of (`str`, `Any`) items
        An optionally weak value dictionary to store objects for assigning them to modules before loading them.
        If it would be set as empty, then it is set as `None` instead.
    _execute_counter : `int`
        Whether the extension loader is executing an extension.
    extensions : `dict` of (`str`, ``Extension``) items
        A dictionary of the added extensions to the extension loader in `extension-name` - ``Extension`` relation.
    
    Class Attributes
    ---------------
    _instance : `None` or ``ExtensionLoader``
        The already created instance of the ``ExtensionLoader`` if there is any.
    """
    __slots__ = ('_default_entry_point', '_default_exit_point', '_default_variables', '_execute_counter',
        'extensions', )
    
    _instance = None
    def __new__(cls,):
        """
        Creates an ``ExtensionLoader`` instance. If the `ExtensionLoader` was instanced already, then returns that
        instead.
        
        Returns
        -------
        self : ``ExtensionLoader``
        """
        self = cls._instance
        if self is None:
            self = object.__new__(cls)
            self._default_entry_point = 'setup'
            self._default_exit_point = 'teardown'
            self.extensions = {}
            self._default_variables = HybridValueDictionary()
            self._execute_counter = 0
        
        return self
    
    @property
    def default_entry_point(self):
        """
        Get-set-del descriptor for modifying the extension loader's default entry point.
        
        Accepts and returns `None`, `str` or a `callable`. If invalid type is given, raises `TypeError`.
        """
        return self._default_entry_point
    
    @default_entry_point.setter
    def default_entry_point(self, default_entry_point):
        if not _validate_entry_or_exit(default_entry_point):
            raise TypeError(f'`{self.__class__.__name__}.default_entry_point` expected `None`, `str` or a `callable`, '
                f'got {default_entry_point.__class__.__name__}.')
        
        self._default_entry_point = default_entry_point
    
    @default_entry_point.deleter
    def default_entry_point(self):
        self._default_entry_point = None
    
    
    @property
    def default_exit_point(self):
        """
        Get-set-del descriptor for modifying the extension loader's default exit point.
        
        Accepts and returns `None`, `str` or a `callable`. If invalid type is given, raises `TypeError`.
        """
        return self._default_exit_point
    
    @default_exit_point.setter
    def default_exit_point(self, default_exit_point):
        if not _validate_entry_or_exit(default_exit_point):
            raise TypeError(f'`{self.__class__.__name__}.default_exit_point` expected `None`, `str` or a `callable`, '
                f'got {default_exit_point.__class__.__name__}.')
        
        self._default_exit_point = default_exit_point
    
    @default_exit_point.deleter
    def default_exit_point(self):
        self._default_exit_point = None
    
    def add_default_variables(self, **variables):
        """
        Adds default variables to the extension loader.
        
        Parameters
        ----------
        **variables : Keyword Arguments
            Variables to assigned to an extension's module before it is loaded.
        
        Raises
        ------
        ValueError
             If a variable name is would be used, what is `module` attribute.
        """
        default_variables = self._default_variables
        for key, value in variables.items():
            if key in PROTECTED_NAMES:
                raise ValueError(f'The passed {key!r} is a protected variable name of module type.')
            default_variables[key] = value
    
    def remove_default_variables(self, *names):
        """
        Removes the mentioned default variables of the extension loader.
        
        If a variable with a specified name is not found, no error is raised.
        
        Parameters
        ----------
        *names : `str`
            Default variable names.
        """
        default_variables = self._default_variables
        for name in names:
            try:
                del default_variables[name]
            except KeyError:
                pass
    
    def clear_default_variables(self):
        """
        Removes all the default variables of the extension loader.
        """
        self._default_variables.clear()
    
    def add(self, name, entry_point=None, exit_point=None, extend_default_variables=True, locked=False,
            take_snapshot_difference=True, **variables):
        """
        Adds an extension to the extension loader.
        
        Parameters
        ----------
        name : `str` or `iterable` of `str`
            The extension(s)'s name(s) to add.
        entry_point : `None`, `str` or `callable`, Optional
            Extension specific entry point, to use over the extension loader's default.
        exit_point : `None`, `str` or `callable`, Optional
            Extension specific exit point, to use over the extension loader's default.
        locked : `bool`, Optional
            Whether the given extension(s) should not be affected by `.{}_all` methods.
        take_snapshot_difference : `bool`, Optional
            Whether snapshot feature should be used.
        **variables : Keyword arguments
            Variables to assign to an extension(s)'s module before they are loaded.
        
        Raises
        ------
        TypeError
            - If `entry_point` was not given as `None`, `str` or as `callable`.
            - If `entry_point` was given as `callable`, but accepts less or more positional arguments, as would be
                given.
            - If `exit_point` was not given as `None`, `str` or as `callable`.
            - If `exit_point` was given as `callable`, but accepts less or more positional arguments, as would be
                given.
            - If `extend_default_variables` was not given as `bool`.
            - If `locked` was not given as `bool`.
            - If `name` was not given as `str` or as `iterable` of `str`.
        ValueError
            If a variable name is would be used, what is `module` attribute.
        """
        if not _validate_entry_or_exit(entry_point):
            raise TypeError(f'`{self.__class__.__name__}.add` expected `None`, `str` or a `callable` as '
                f'`entry_point`, got {entry_point.__class__.__name__}.')
        
        if not _validate_entry_or_exit(exit_point):
            raise TypeError(f'`{self.__class__.__name__}.add` expected `None`, `str` or a `callable` as `exit_point`, '
                f'got {exit_point.__class__.__name__}.')
        
        if variables:
            default_variables = HybridValueDictionary(variables)
            for key, value in variables.items():
                if key in PROTECTED_NAMES:
                    raise ValueError(f'The passed {key!r} is a protected variable name of module type.')
                default_variables[key] = value
        else:
            default_variables = None
        
        extend_default_variables_type = extend_default_variables.__class__
        if extend_default_variables_type is bool:
            pass
        elif issubclass(extend_default_variables_type, int):
            extend_default_variables = bool(extend_default_variables)
        else:
            raise TypeError(f'`extend_default_variables` should have been passed as `bool`, got: '
                f'{extend_default_variables_type.__name__}.')
        
        locked_type = locked.__class__
        if locked_type is bool:
            pass
        elif issubclass(locked_type, int):
            locked = bool(locked)
        else:
            raise TypeError(f'`locked` should have been passed as `bool`, got: {locked_type.__name__}.')
        
        name_type = name.__class__
        if name_type is str:
            pass
        
        elif issubclass(name_type, str):
            name = str(name)
        
        if hasattr(name_type, '__iter__'):
            names = []
            index = 1
            for name_ in name:
                name_type = name_.__class__
                if name_type is str:
                    pass
                elif issubclass(name_type, str):
                    name = str(name)
                else:
                    raise TypeError(f'`{self.__class__.__name__}.add` expected `str` or `iterable` of `str` as '
                        f'`name`, got, `iterable`, but it\'s {index}. element is not `str` instance, got: '
                        f'{name_type.__class__}')
                
                names.append(name)
                index += 1
            
            for name in names:
                extension = Extension(name, entry_point, exit_point, extend_default_variables, locked,
                    take_snapshot_difference, default_variables)
                
                self.extensions[name] = extension
                
                dot_index = name.rfind('.')
                if dot_index != -1:
                    name = name[dot_index+1:]
                    
                    self.extensions.setdefault(name, extension)
            
            return
        else:
            raise TypeError(f'`{self.__class__.__name__}.add` expected `str` or `iterable of str` as `name`, got '
                f'`{name_type.__name__}`.')
        
        extension = Extension(name, entry_point, exit_point, extend_default_variables, locked, take_snapshot_difference,
            default_variables)
        
        self.extensions[name] = extension
        
        dot_index = name.rfind('.')
        if dot_index != -1:
            name = name[dot_index+1:]
            
            self.extensions.setdefault(name, extension)
        
    def remove(self, name):
        """
        Removes one or more extensions from the extension loader.
        
        If any of the extensions is not found, no errors will be raised.
        
        Parameters
        ----------
        name : `str` or `iterable` of `str`
            The extension(s)'s name(s) to remove.
        
        Raises
        ------
        TypeError
            If `name` was not given as `str` or as `iterable` of `str`.
        RuntimeError
            If a loaded extension would be removed.
        """
        name_type = name.__class__
        if name_type is str:
            pass
        
        elif issubclass(name_type, str):
            name = str(name)
        
        elif hasattr(name_type, '__iter__'):
            names = []
            index = 1
            for name in name:
                name_type = name.__class__
                if name_type is str:
                    pass
                elif issubclass(name_type, str):
                    name = str(name)
                else:
                    raise TypeError(f'`{self.__class__.__name__}.remove` expected `str` or `iterable` of `str` as '
                        f'`name`, got, `iterable`, but it\'s {index}. element is not `str` instance, got: '
                        f'{name_type.__class__}')
                
                names.append(name)
                index += 1
            
            collected = []
            extensions = self.extensions
            for name in names:
                try:
                    extension = extensions[name]
                except KeyError:
                    continue
                
                if extension._state == EXTENSION_STATE_LOADED:
                    raise RuntimeError(f'Extension `{name!r}` can not be removed, meanwhile it is loaded.')
                
                collected.append(extension)
                
            for extension in collected:
                extension._unlink()
            
            return
        else:
            #no more good case -> raise
            raise TypeError(f'`{self.__class__.__name__}.remove` expected `str` or `iterable of str` as `name`, got '
                f'{name_type.__name__}.')

        try:
            extension = self.extensions[name]
        except KeyError:
            return
        
        if extension._state == EXTENSION_STATE_LOADED:
            raise RuntimeError(f'Extension `{name!r}` can not be removed, meanwhile it is loaded.')
        
        extension._unlink()
        return
    
    def load_extension(self, name, *args, **kwargs):
        """
        Adds, then loads the extension.
        
        Parameters
        ----------
        name : `str`
            The extension's name.
        *args : Argument
            Additional argument to create the extension with.
        **kwargs : Keyword Arguments
            Additional keyword arguments to create the extension with.
        
        Other Parameters
        ----------------
        entry_point : `None, `str` or `callable`, Optional (Keyword only)
            Extension specific entry point, to use over the extension loader's default.
        exit_point : `None, `str` or `callable`, Optional (Keyword only)
            Extension specific exit point, to use over the extension loader's default.
        locked : `bool`, Optional (Keyword only)
            Whether the given extension(s) should not be affected by `.{}_all` methods.
        take_snapshot_difference : `bool`, Optional (Keyword only)
            Whether snapshot feature should be used.
        **variables : Keyword arguments
            Variables to assign to an extension(s)'s module before they are loaded.
        
        Returns
        -------
        task : `None`, ``Task``, ``FutureAsyncWrapper``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            loading is done. However if called from a sync thread, will block till the loading is done.
        
        Raises
        ------
        TypeError
            - If `name` was not given as `str`.
            - If `entry_point` was not given as `None`, `str` or as `callable`.
            - If `entry_point` was given as `callable`, but accepts less or more positional arguments, as would be
                given.
            - If `exit_point` was not given as `None`, `str` or as `callable`.
            - If `exit_point` was given as `callable`, but accepts less or more positional arguments, as would be
                given.
            - If `extend_default_variables` was not given as `bool`.
            - If `locked` was not given as `bool`.
        ValueError
            If a variable name is would be used, what is `module` attribute.
        ExtensionError
            The extension failed to load correctly.
        """
        name_type = name.__class__
        if name_type is str:
            pass
        elif issubclass(name_type, str):
            name = str(name)
        else:
            raise TypeError(f'`name` should have be given as `str` instance, got {name_type.__name__}.')
        
        self.add(name, *args, **kwargs)
        return self.load(name)
    
    def load(self, name):
        """
        Loads the extension with the given name. If the extension is already loaded, will do nothing.
        
        Returns
        -------
        task : `None`, ``Task``, ``FutureAsyncWrapper``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            loading is done. However if called from a sync thread, will block till the loading is done.
        
        Raises
        ------
        ExtensionError
            - No extension is added with the given name.
            - Loading the extension failed.
        """
        task = Task(self._load(name), KOKORO)
        
        thread = current_thread()
        if thread is KOKORO:
            return task
        
        if isinstance(current_thread, EventThread):
            return task.async_wrap(current_thread)
        
        KOKORO.wake_up()
        task.sync_wrap().wait()
    
    async def _load(self, name):
        """
        Loads the extension with the given name.
        
        This method is a coroutine.
        
        Raises
        ------
        ExtensionError
            - No extension is added with the given name.
            - Loading the extension failed.
        """
        try:
            extension = self.extensions[name]
        except KeyError:
            raise ExtensionError(f'No extension was added with name: `{name}`.') from None
        
        await self._load_extension(extension)
        
    def unload(self, name):
        """
        Unloads the extension with the given name.
        
        Returns
        -------
        task : `None`, ``Task``, ``FutureAsyncWrapper``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            unloading is done. However if called from a sync thread, will block till the unloading is done.
        
        Raises
        ------
        ExtensionError
            - No extension is added with the given name.
            - Unloading the extension failed.
        """
        task = Task(self._unload(name), KOKORO)
        
        thread = current_thread()
        if thread is KOKORO:
            return task
        
        if isinstance(current_thread, EventThread):
            return task.async_wrap(current_thread)
        
        KOKORO.wake_up()
        task.sync_wrap().wait()
    
    async def _unload(self, name):
        """
        Unloads the extension with the given name.
        
        This method is a coroutine.
        
        Raises
        ------
        ExtensionError
            - No extension is added with the given name.
            - Unloading the extension failed.
        """
        try:
            extension = self.extensions[name]
        except KeyError:
            raise ExtensionError(f'No extension was added with name: `{name}`.') from None
        
        await self._unload_extension(extension)
        
    def reload(self, name):
        """
        Reloads the extension with the given name.
        
        Returns
        -------
        task : `None`, ``Task``, ``FutureAsyncWrapper``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            reloading is done. However if called from a sync thread, will block till the reloading is done.
        
        Raises
        ------
        ExtensionError
            - No extension is added with the given name.
            - Reloading the extension failed.
        """
        task = Task(self._reload(name),KOKORO)
        
        thread = current_thread()
        if thread is KOKORO:
            return task
        
        if isinstance(current_thread, EventThread):
            return task.async_wrap(current_thread)
        
        KOKORO.wake_up()
        task.sync_wrap().wait()
    
    async def _reload(self, name):
        """
        Reloads the extension with the given name.
        
        This method is a coroutine.
        
        Raises
        ------
        ExtensionError
            - No extension is added with the given name.
            - Reloading the extension failed.
        """
        try:
            extension = self.extensions[name]
        except KeyError:
            raise ExtensionError(f'No extension was added with name: `{name}`.') from None
        
        await self._unload_extension(extension)
        await self._load_extension(extension)
    
    def load_all(self):
        """
        Loads all the extension of the extension loader. If anything goes wrong, raises an ``ExtensionError`` only
        at the end, with the exception(s).
        
        Returns
        -------
        task : `None`, ``Task``, ``FutureAsyncWrapper``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            loading is done. However if called from a sync thread, will block till the loading is done.
        
        Raises
        ------
        ExtensionError
            If any extension failed to load correctly.
        """
        task = Task(self._load_all(),KOKORO)

        thread = current_thread()
        if thread is KOKORO:
            return task
        
        if isinstance(current_thread, EventThread):
            return task.async_wrap(current_thread)
        
        KOKORO.wake_up()
        task.sync_wrap().wait()
        
    async def _load_all(self):
        """
        Loads all the extensions of the extension loader.
        
        Loads each extension one after the other. The raised exceptions' messages are collected into one exception,
        what will be raised only at the end. If any of the extensions raises, will still try to unload the leftover
        ones.
        
        This method is a coroutine.
        
        Raises
        ------
        ExtensionError
            If any extension failed to load correctly.
        """
        error_messages = []
        
        for extension in self.extensions.values():
            if extension._locked:
                continue
            
            try:
                await self._load_extension(extension)
            except ExtensionError as err:
                error_messages.append(err.message)
            
        if error_messages:
            raise ExtensionError(error_messages) from None
        
    def unload_all(self):
        """
        Unloads all the extension of the extension loader. If anything goes wrong, raises an ``ExtensionError`` only
        at the end, with the exception(s).
        
        Returns
        -------
        task : `None`, ``Task``, ``FutureAsyncWrapper``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            unloading is done. However if called from a sync thread, will block till the unloading is done.
        
        Raises
        ------
        ExtensionError
            If any extension failed to unload correctly.
        """
        task = Task(self._unload_all(), KOKORO)
        
        thread = current_thread()
        if thread is KOKORO:
            return task
        
        if isinstance(current_thread, EventThread):
            return task.async_wrap(current_thread)
        
        KOKORO.wake_up()
        task.sync_wrap().wait()
    
    async def _unload_all(self):
        """
        Unloads all the extensions of the extension loader.
        
        Unloads each extension one after the other. The raised exceptions' messages are collected into one exception,
        what will be raised only at the end. If any of the extensions raises, will still try to unload the leftover
        ones.
        
        This method is a coroutine.
        
        Raises
        ------
        ExtensionError
            If any extension failed to unload correctly.
        """
        error_messages = []
        
        for extension in self.extensions.values():
            if extension._locked:
                continue
            
            try:
                await self._unload_extension(extension)
            except ExtensionError as err:
                error_messages.append(err.message)
            
        if error_messages:
            raise ExtensionError(error_messages) from None
    
    def reload_all(self):
        """
        Reloads all the extension of the extension loader. If anything goes wrong, raises an ``ExtensionError`` only
        at the end, with the exception(s).
        
        Returns
        -------
        task : `None`, ``Task``, ``FutureAsyncWrapper``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            reloading is done. However if called from a sync thread, will block till the reloading is done.
        
        Raises
        ------
        ExtensionError
            If any extension failed to reload correctly.
        """
        task = Task(self._reload_all(), KOKORO)
        
        thread = current_thread()
        if thread is KOKORO:
            return task
        
        if isinstance(current_thread, EventThread):
            return task.async_wrap(current_thread)
        
        KOKORO.wake_up()
        task.sync_wrap().wait()
    
    async def _reload_all(self):
        """
        Reloads all the extensions of the extension loader.
        
        If an extension is not unloaded, will load it, if the extension is loaded, will unload, then load it.
        Reloads each extension one after the other. The raised exceptions' messages are collected into one exception,
        what will be raised at the end. If any of the extensions raises, will still try to reload the leftover ones.
        
        This method is a coroutine.
        
        Raises
        ------
        ExtensionError
            If any extension failed to reload correctly.
        """
        error_messages = []
        
        for extension in self.extensions.values():
            if extension._locked:
                continue
            
            try:
                await self._unload_extension(extension)
            except ExtensionError as err:
                error_messages.append(err.message)
                continue
            
            try:
                await self._load_extension(extension)
            except ExtensionError as err:
                error_messages.append(err.message)
        
        if error_messages:
            raise ExtensionError(error_messages) from None
    
    async def _load_extension(self, extension):
        """
        Loads the extension. If the extension is loaded, will do nothing.
        
        Loading an exception can be separated to 4 parts:
        
        - Assign the default variables.
        - Load the module.
        - Find the entry point (if needed).
        - Ensure the entry point (if found).
        
        If any of these fails, an ``ExtensionError`` will be raised. If step 1 raises, then a traceback will be
        included as well.
        
        This method is a coroutine.
        
        Parameters
        ----------
        extension : ``Extension``
            The extension to load.
        
        Raises
        ------
        ExtensionError
            Extension entry point raised.
        """
        self._execute_counter += 1
        try:
            try:
                # loading blocks, but unloading does not
                lib = await KOKORO.run_in_executor(extension._load)
            except BaseException as err:
                message = await KOKORO.run_in_executor(alchemy_incendiary(
                    self._render_exc, (err, [
                    'Exception occurred meanwhile loading an extension: `', extension.name, '`.\n\n',],
                        )))
                
                raise ExtensionError(message) from None
            
            if lib is None:
                return # already loaded
            
            entry_point = extension._entry_point
            if entry_point is None:
                entry_point = self._default_entry_point
                if entry_point is None:
                    return
            
            if isinstance(entry_point, str):
                entry_point = getattr(lib, entry_point, None)
                if entry_point is None:
                    return
            
            try:
                
                if is_coro(entry_point):
                    await entry_point(lib)
                else:
                    entry_point(lib)
            
            except BaseException as err:
                message = await KOKORO.run_in_executor(alchemy_incendiary(
                    self._render_exc, (err, [
                    'Exception occurred meanwhile entering an extension: `', extension.name,
                    '`.\nAt entry_point:', repr(entry_point), '\n\n',],
                        )))
                
                raise ExtensionError(message) from None
        finally:
            self._execute_counter -= 1
        
    async def _unload_extension(self, extension):
        """
        Unloads the extension. If the extension is not loaded, will do nothing.
        
        Loading an exception can be separated to 3 parts:
        
        - Find the exit point (if needed).
        - Ensure the exit point (if found).
        - Remove the default variables.
        
        If any of these fails, an ``ExtensionError`` will be raised. If step 2 raises, then a traceback will be
        included as well.
        
        This method is a coroutine.
        
        Parameters
        ----------
        extension : ``Extension``
            The extension to unload.
        
        Raises
        ------
        ExtensionError
            Extension exit point raised.
        """
        self._execute_counter += 1
        try:
            # loading blocks, but unloading does not
            lib = extension._unload()
            
            if lib is None:
                return # not loaded
            
            try:
                exit_point = extension._exit_point
                if exit_point is None:
                    exit_point = self._default_exit_point
                    if exit_point is None:
                        return
                
                if isinstance(exit_point, str):
                    exit_point = getattr(lib, exit_point, None)
                    if exit_point is None:
                        return
                
                try:
                    
                    if is_coro(exit_point):
                        await exit_point(lib)
                    else:
                        exit_point(lib)
                
                except BaseException as err:
                    message = await KOKORO.run_in_executor(alchemy_incendiary(
                        self._render_exc, (err, [
                        'Exception occurred meanwhile unloading an extension: `', extension.name,
                        '`.\nAt exit_point:', repr(exit_point), '\n\n',],
                            )))
                    
                    raise ExtensionError(message) from None
            
            finally:
                extension._unassign_variables()
                
                keys = []
                lib_globals = lib.__dict__
                for key in lib_globals:
                    if (not key.startswith('__')) and (not key.endswith('__')):
                        keys.append(key)
                
                for key in keys:
                    del lib_globals[key]
        finally:
            self._execute_counter -= 1
    
    @staticmethod
    def _render_exc(exception, header):
        """
        Renders an exception' traceback.
        
        Parameters
        ----------
        exception : `BaseException` instance
            The exception to render.
        header : `str` or `list` of `str`
            Header of the rendered traceback.
        
        Returns
        -------
        message : `str`
            The `exception`'s rendered traceback.
        
        Notes
        -----
        This function should run in an executor.
        """
        file = StringIO()
        EventThread._render_exc_sync(exception, before=header, after=None, file=file)
        message = file.getvalue()
        file.close()
        
        return message
    
    def __repr__(self):
        """Returns the extension loader's representation."""
        result = [
            '<',
            self.__class__.__name__,
            ' extension count=',
            repr(len(self.extensions)),
        ]
        
        entry_point = self._default_entry_point
        if (entry_point is not None):
            result.append(', default_entry_point=')
            result.append(repr(entry_point))
        
        exit_point = self._default_exit_point
        if (exit_point is not None):
            result.append(', default_exit_point=')
            result.append(repr(exit_point))
        
        default_variables = self._default_variables
        if default_variables:
            result.append(', default_variables=')
            result.append(repr(default_variables))
        
        result.append('>')
        
        return ''.join(result)

    def is_processing_extension(self):
        """
        Returns whether the extension loader is processing an extension.
        
        Returns
        -------
        is_processing_extension : `bool`
        """
        if self._execute_counter:
            is_processing_extension = True
        else:
            is_processing_extension = False
        
        return is_processing_extension


EXTENSION_LOADER = ExtensionLoader()



# -*- coding: utf-8 -*-
from .extension_loader import *

__all__ = extension_loader.__all__

from .. import register_library_extension
register_library_extension('HuyaneMatsu.extension_loader')
del register_library_extension



# -*- coding: utf-8 -*-
from .color import Color

def preconvert_snowflake(snowflake, name):
    """
    Converts the given `snowflake` to an acceptable value by the wrapper.
    
    Parameters
    ----------
    snowflake : `str` or `int`
        The snowflake to convert.
    name : `str`
        The name of the snowflake.
    
    Returns
    -------
    snowflake : `int`
    
    Raises
    ------
    TypeError
        - If `snowflake` was not passed neither as `int` or as `str` instance.
    ValueError
        - If `snowflake` was passed as `str` and cannot be converted to `int`.
        - If the converted `snowflake` is negative or it's bit length is over 64.
    """
    snowflake_type = snowflake.__class__
    if snowflake_type is int:
        pass
    if issubclass(snowflake_type, int):
        snowflake = int(snowflake)
    # JSON uint64 is str
    elif issubclass(snowflake_type, str):
        if 6 < len(snowflake) < 21 and snowflake.isdigit():
            snowflake = int(snowflake)
        else:
            raise ValueError(f'`{name}` can be passed as `int` or `str` instance, got `str` instance, but not a valid'
                f'snowflake (7-20 length, digit only), got {snowflake!r}.')
    else:
        raise TypeError(f'`{name}` can be passed as `int` or `str` instance, got {snowflake_type.__name__}.')
    
    if snowflake < 0 or snowflake > ((1<<64)-1):
        raise ValueError(f'`{name}` can be only uint64, got {snowflake!r}.')
    
    return snowflake

def preconvert_discriminator(discriminator):
    """
    Converts the given `discriminator` to an acceptable value by the wrapper.
    
    Parameters
    ----------
    discriminator : `str` or `int`
        The discriminator of an user to convert.
    
    Returns
    -------
    discriminator : `int`
    
    Raises
    ------
    TypeError
        If `discriminator` was not passed neither as `int` or as `str` instance.
    ValueError
        - If `discriminator` was passed as `str` and it is not numerical or it's length is over `4`.
        - If the `discriminator`'s value is less than `0` or is over than `9999`.
    """
    if (type(discriminator) is int):
        pass
    elif isinstance(discriminator, int):
        discriminator = int(discriminator)
    # Discord sends `discriminator` as `str`, so lets accept that as well.
    elif isinstance(discriminator, str):
        if 0<len(discriminator) < 5 and discriminator.isdigit():
            raise ValueError(f'`discriminator` was given as a `str` instance, but it is not numerical or it\'s length '
                 f'is over `4`, got {discriminator!r}')
        discriminator = int(discriminator)
    else:
        raise TypeError(f'`discriminator` can be passed as `int` or `str` instance, got '
            f'{discriminator.__class__.__name__}.')
    
    if discriminator < 0 or discriminator > 9999:
        raise ValueError(f'`discriminator` can be between 0 and 9999, got got {discriminator!r}.')
    
    return discriminator

def preconvert_color(color):
    """
    Converts the given `color` to an acceptable value by the wrapper.
    
    Parameters
    ----------
    color : `int`
        The color to convert.
    
    Returns
    -------
    color : `Color`
    
    Raises
    ------
    TypeError
        If `color` was not passed as `int` instance.
    ValueError
        If the `color`'s value is less than `0` or is over than `0xffffff`.
    """
    if type(color) is Color:
        pass
    elif isinstance(color, int):
        color = Color(color)
    else:
        raise TypeError(f'`color` can be `{Color.__name__}` or `int` instance, got {color.__class__.__name__}.')
    
    if color < 0 or color > 0xffffff:
        raise ValueError(f'`color` can be between 0 and 0xffffff, got {color!r}.')
    
    return color

def preconvert_str(value, name, lower_limit, upper_limit):
    """
    Converts the given `value` to an acceptable string by the wrapper.
    
    Parameters
    ----------
    value : `str`
        The string to convert,
    name : `str`
        The name of the value.
    lower_limit : `int`
        The minimal length of the string.
    upper_limit : `int`
        The maximal length of the string.
    
    Returns
    -------
    value : `str`
    
    Raises
    ------
    TypeError
        If `value` was not passed as `str` instance.
    ValueError
        If the `value`'s length is less than the given `lower_limit` or is higher than the given than the given
        `upper_limit`.
    """
    if type(value) is str:
        pass
    elif isinstance(value, str):
        value = str(value)
    else:
        raise TypeError(f'`{name}` can be `str` instance, got {value.__class__.__name__}.')
    
    length = len(value)
    if (length != 0) and (length < lower_limit or length > upper_limit):
        raise ValueError(f'`{name}` can be between length {lower_limit} and {upper_limit}, got {value!r}.')
    
    return value

def preconvert_bool(value, name):
    """
    Converts the given `value` to an acceptable boolean by the wrapper.
    
    Parameters
    ----------
    value : `int`
        The value to convert.
    name : `str`
        The name of the value.
    
    Returns
    -------
    value : `str`
    
    Raises
    ------
    TypeError
        If `value` was not passed as `int` instance.
    ValueError
        If value was passed as an `int` instance, but not as `0` or `1` either.
    """
    if (type(value) is bool):
        pass
    elif isinstance(value, int):
        if (value not in (0, 1)):
            raise ValueError(f'`{name}` was given as `int` instance, but neither as `0` or `1`, got {value!r}.')
        value = bool(value)
    else:
        raise TypeError(f'`{name}` can be `bool` or `int` instance as `0` or `1`, got {value.__class__.__name__}.')
    
    return value

def preconvert_flag(flag, name, type_):
    """
    Converts the given `flag` to an acceptable flag by the wrapper.
    
    Parameters
    ----------
    flag : `int`
        The flag to convert.
    name : `str`
        The name of the flag.
    type_ : `int` subtype
        The type of the output flag.
    
    Returns
    -------
    flag : `type_`
    
    Raises
    ------
    TypeError
        If `flag` was not given as `int` instance.
    ValueError
        If `flag` was given as `int` instance, but it's value is less than `0` or it's bit length is over `64`.
    """
    flag_type = flag.__class__
    if (flag_type is type_):
        pass
    elif issubclass(flag_type, int):
        flag = type_(flag)
    else:
        raise TypeError(f'`{name}` can be passed as `{type_.__name__}` instance, got {flag_type.__name__}.')
    
    if flag < 0 or flag > ((1<<64)-1):
        raise ValueError(f'`{name}` can be only uint64, got {flag!r}.')
    
    return flag

def preconvert_preinstanced_type(value, name, type_):
    """
    Converts the given `value` to an acceptable value by the wrapper.
    
    Parameters
    ----------
    value : `Any`
        The value to convert.
    name : `str`
        The name of the value.
    type_ : ``PreinstancedBase`` instance
        The preinstanced type.
    
    Returns
    -------
    value : ``PreinstancedBase`` instance
    
    Raises
    ------
    TypeError
        If `value` was not given as `type_` instance, neither as `type_.value`'s type's instance.
    ValueError
        If there is no preinstanced object for the given `value`.
    """
    value_type = value.__class__
    if (value_type is not type_):
        value_expected_type = type_.VALUE_TYPE
        if value_type is value_expected_type:
            pass
        elif issubclass(value_type, value_expected_type):
            value = value_expected_type(value)
        else:
            raise TypeError(f'`{name}` can be passed as {type_.__name__} or as {value_expected_type.__name__} '
                f'instance, got {value_type.__name__}.')
        
        try:
            value = type_.INSTANCES[value]
        except LookupError:
            raise ValueError(f'There is no predefined `{name}` for the following value: {value!r}.') from None
    
    return value

def preconvert_int(value, name, lower_limit, upper_limit):
    """
    Converts the given `value` to an acceptable integer by the wrapper.
    
    Parameters
    ----------
    value : `Any`
        The value to convert.
    name : `str`
        The name of the value.
    lower_limit : `int`
        The minimal value of `value`.
    upper_limit : `int`
        The maximal value of `value`.
    
    Returns
    -------
    value : `int`
    
    Raises
    ------
    TypeError
        If `value` was not given as `int` instance.
    ValueError
        If `value` is less than `lower_limit`, or is higher than the `upper_limit`.
    """
    if type(value) is int:
        pass
    elif isinstance(value, int):
        value = int(value)
    else:
        raise TypeError(f'`{name}` can be `int` instance, got {value.__class__.__name__}.')
    
    if value < lower_limit or value > upper_limit:
        raise ValueError(f'`{name}` can be between {lower_limit} and {upper_limit}, got {value!r}.')
    
    return value
